@@ -1,41 +1,75 @@
 """Provides commands and queries for Git."""
-import os
+from __future__ import division, absolute_import, unicode_literals
+
 import re
-from cStringIO import StringIO
+from io import StringIO
 
 from cola import core
-from cola import git
 from cola import gitcfg
-from cola import errors
 from cola import utils
 from cola import version
-from cola.compat import set
-from cola.decorators import memoize
+from cola.git import git
+from cola.git import STDOUT
+from cola.i18n import N_
 
-git = git.instance()
-config = gitcfg.instance()
 
-class InvalidRepositoryError(StandardError):
+class InvalidRepositoryError(Exception):
     pass
 
 
-def default_remote():
+def default_remote(config=None):
     """Return the remote tracked by the current branch."""
+    if config is None:
+        config = gitcfg.current()
     return config.get('branch.%s.remote' % current_branch())
 
 
-def diff_filenames(arg, git=git):
+def diff_index_filenames(ref):
+    """Return a of filenames that have been modified relative to the index"""
+    out = git.diff_index(ref, name_only=True, z=True)[STDOUT]
+    return _parse_diff_filenames(out)
+
+
+def diff_filenames(*args):
     """Return a list of filenames that have been modified"""
-    diff_zstr = git.diff(arg, name_only=True, z=True,
-                         **_common_diff_opts()).rstrip('\0')
-    return [core.decode(f) for f in diff_zstr.split('\0') if f]
+    out = git.diff_tree(name_only=True, no_commit_id=True, r=True, z=True,
+                        *args)[STDOUT]
+    return _parse_diff_filenames(out)
 
 
-def all_files(git=git):
+def diff(args):
+    """Return a list of filenames for the given diff arguments
+
+    :param args: list of arguments to pass to "git diff --name-only"
+
+    """
+    out = git.diff(name_only=True, z=True, *args)[STDOUT]
+    return _parse_diff_filenames(out)
+
+
+def _parse_diff_filenames(out):
+    if out:
+        return out[:-1].split('\0')
+    else:
+        return []
+
+
+def tracked_files():
     """Return the names of all files in the repository"""
-    return [core.decode(f)
-            for f in git.ls_files(z=True)
-                        .strip('\0').split('\0') if f]
+    out = git.ls_files(z=True)[STDOUT]
+    if out:
+        return sorted(out[:-1].split('\0'))
+    else:
+        return []
+
+
+def all_files():
+    """Returns a sorted list of all files, including untracked files."""
+    ls_files = git.ls_files(z=True,
+                            cached=True,
+                            others=True,
+                            exclude_standard=True)[STDOUT]
+    return sorted([f for f in ls_files.split('\0') if f])
 
 
 class _current_branch:
@@ -44,27 +78,25 @@ class _current_branch:
     value = None
 
 
-def clear_cache():
+def reset():
     _current_branch.key = None
 
 
-def current_branch(git=git):
+def current_branch():
     """Return the current branch"""
     head = git.git_path('HEAD')
     try:
-        key = os.stat(head).st_mtime
+        key = core.stat(head).st_mtime
         if _current_branch.key == key:
             return _current_branch.value
-    except OSError, e:
+    except OSError:
         pass
-    data = git.rev_parse('HEAD', with_stderr=True, symbolic_full_name=True)
-    if data.startswith('fatal:'):
-        # git init -- read .git/HEAD.  We could do this unconditionally
-        # and avoid the subprocess call.  It's probably time to start
-        # using dulwich.
+    status, data, err = git.rev_parse('HEAD', symbolic_full_name=True)
+    if status != 0:
+        # git init -- read .git/HEAD.  We could do this unconditionally...
         data = _read_git_head(head)
 
-    for refs_prefix in ('refs/heads/', 'refs/remotes/'):
+    for refs_prefix in ('refs/heads/', 'refs/remotes/', 'refs/tags/'):
         if data.startswith(refs_prefix):
             value = data[len(refs_prefix):]
             _current_branch.key = key
@@ -77,15 +109,15 @@ def current_branch(git=git):
 def _read_git_head(head, default='master', git=git):
     """Pure-python .git/HEAD reader"""
     # Legacy .git/HEAD symlinks
-    if os.path.islink(head):
-        refs_heads = os.path.realpath(git.git_path('refs', 'heads'))
-        path = os.path.abspath(head).replace('\\', '/')
+    if core.islink(head):
+        refs_heads = core.realpath(git.git_path('refs', 'heads'))
+        path = core.abspath(head).replace('\\', '/')
         if path.startswith(refs_heads + '/'):
             return path[len(refs_heads)+1:]
 
     # Common .git/HEAD "ref: refs/heads/master" file
-    elif os.path.isfile(head):
-        data = utils.slurp(head).rstrip()
+    elif core.isfile(head):
+        data = core.read(head).rstrip()
         ref_prefix = 'ref: '
         if data.startswith(ref_prefix):
             return data[len(ref_prefix):]
@@ -110,9 +142,10 @@ def branch_list(remote=False):
 
 def for_each_ref_basename(refs, git=git):
     """Return refs starting with 'refs'."""
-    output = git.for_each_ref(refs, format='%(refname)').splitlines()
+    out = git.for_each_ref(refs, format='%(refname)')[STDOUT]
+    output = out.splitlines()
     non_heads = filter(lambda x: not x.endswith('/HEAD'), output)
-    return map(lambda x: x[len(refs) + 1:], non_heads)
+    return list(map(lambda x: x[len(refs) + 1:], non_heads))
 
 
 def all_refs(split=False, git=git):
@@ -124,7 +157,8 @@ def all_refs(split=False, git=git):
     query = (triple('refs/tags', tags),
              triple('refs/heads', local_branches),
              triple('refs/remotes', remote_branches))
-    for ref in git.for_each_ref(format='%(refname)').splitlines():
+    out = git.for_each_ref(format='%(refname)')[STDOUT]
+    for ref in out.splitlines():
         for prefix, prefix_len, dst in query:
             if ref.startswith(prefix) and not ref.endswith('/HEAD'):
                 dst.append(ref[prefix_len:])
@@ -135,10 +169,14 @@ def all_refs(split=False, git=git):
         return local_branches + remote_branches + tags
 
 
-def tracked_branch(branch=None):
+def tracked_branch(branch=None, config=None):
     """Return the remote branch associated with 'branch'."""
+    if config is None:
+        config = gitcfg.current()
     if branch is None:
         branch = current_branch()
+    if branch is None:
+        return None
     remote = config.get('branch.%s.remote' % branch)
     if not remote:
         return None
@@ -151,66 +189,95 @@ def tracked_branch(branch=None):
     return None
 
 
-def untracked_files(git=git):
+def untracked_files(git=git, paths=None):
     """Returns a sorted list of untracked files."""
-    ls_files = git.ls_files(z=True,
-                            others=True,
-                            exclude_standard=True)
-    return [core.decode(f) for f in ls_files.split('\0') if f]
+
+    if paths is None:
+        paths = []
+    filter_paths = ['--'] + paths
+    out = git.ls_files(z=True, others=True, exclude_standard=True,
+                       *filter_paths)[STDOUT]
+    if out:
+        return out[:-1].split('\0')
+    return []
 
 
 def tag_list():
     """Return a list of tags."""
-    tags = for_each_ref_basename('refs/tags')
-    tags.reverse()
-    return tags
+    return list(reversed(for_each_ref_basename('refs/tags')))
+
+
+def log(git, *args, **kwargs):
+    return git.log(no_color=True, no_ext_diff=True, *args, **kwargs)[STDOUT]
 
 
 def commit_diff(sha1, git=git):
-    commit = git.show(sha1)
-    first_newline = commit.index('\n')
-    if commit[first_newline+1:].startswith('Merge:'):
-        return (core.decode(commit) + '\n\n' +
-                core.decode(diff_helper(commit=sha1,
-                                             cached=False,
-                                             suppress_header=False)))
-    else:
-        return core.decode(commit)
+    return log(git, '-1', sha1, '--') + '\n\n' + sha1_diff(git, sha1)
 
 
-def _common_diff_opts(config=config):
-    patience = version.check('patience', version.git_version())
+_diff_overrides = {}
+def update_diff_overrides(space_at_eol, space_change,
+                          all_space, function_context):
+    _diff_overrides['ignore_space_at_eol'] = space_at_eol
+    _diff_overrides['ignore_space_change'] = space_change
+    _diff_overrides['ignore_all_space'] = all_space
+    _diff_overrides['function_context'] = function_context
+
+
+def common_diff_opts(config=None):
+    if config is None:
+        config = gitcfg.current()
     submodule = version.check('diff-submodule', version.git_version())
-    return {
-        'patience': patience,
+    opts = {
+        'patience': True,
         'submodule': submodule,
         'no_color': True,
         'no_ext_diff': True,
-        'with_raw_output': True,
-        'with_stderr': True,
         'unified': config.get('gui.diffcontext', 3),
+        '_raw': True,
     }
+    opts.update(_diff_overrides)
+    return opts
 
 
-def sha1_diff(sha1, git=git):
-    return git.diff(sha1 + '^!', **_common_diff_opts())
+def _add_filename(args, filename):
+    if filename:
+        args.extend(['--', filename])
 
 
-def diff_info(sha1, git=git):
-    log = git.log('-1',
-                  '--format=Author:\t%aN <%aE>%n'
-                  'Date:\t%aD%n%n'
-                  '%s%n%n%b',
-                  sha1)
-    return log + '\n\n' + sha1_diff(sha1)
+def sha1_diff(git, sha1, filename=None):
+    """Return the diff for a sha1"""
+    # Naively "$sha1^!" is what we'd like to use but that doesn't
+    # give the correct result for merges--the diff is reversed.
+    # Be explicit and compare sha1 against its first parent.
+    args = [sha1 + '~', sha1]
+    opts = common_diff_opts()
+    _add_filename(args, filename)
+    status, out, err = git.diff(*args, **opts)
+    if status != 0:
+        # We probably don't have "$sha1~" because this is the root commit.
+        # "git show" is clever enough to handle the root commit.
+        args = [sha1 + '^!']
+        _add_filename(args, filename)
+        status, out, err = git.show(pretty='format:', *args, **opts)
+        out = out.lstrip()
+    return out
+
+
+def diff_info(sha1, git=git, filename=None):
+    decoded = log(git, '-1', sha1, '--', pretty='format:%b').strip()
+    if decoded:
+        decoded += '\n\n'
+    return decoded + sha1_diff(git, sha1, filename=filename)
 
 
 def diff_helper(commit=None,
-                branch=None,
                 ref=None,
                 endref=None,
                 filename=None,
                 cached=True,
+                head=None,
+                amending=False,
                 with_diff_header=False,
                 suppress_header=True,
                 reverse=False,
@@ -222,59 +289,68 @@ def diff_helper(commit=None,
     if ref and endref:
         argv.append('%s..%s' % (ref, endref))
     elif ref:
-        for r in ref.strip().split():
+        for r in utils.shell_split(ref.strip()):
             argv.append(r)
-    elif branch:
-        argv.append(branch)
+    elif head and amending and cached:
+        argv.append(head)
 
+    encoding = None
     if filename:
         argv.append('--')
         if type(filename) is list:
             argv.extend(filename)
         else:
             argv.append(filename)
+            cfg = gitcfg.current()
+            encoding = cfg.file_encoding(filename)
 
-    start = False
-    del_tag = 'deleted file mode '
+    if filename is not None:
+        deleted = cached and not core.exists(filename)
+    else:
+        deleted = False
 
-    headers = []
-    deleted = cached and not os.path.exists(core.encode(filename))
-
-    # The '--patience' option did not appear until git 1.6.2
-    # so don't allow it to be used on version previous to that
-    patience = version.check('patience', version.git_version())
-    submodule = version.check('diff-submodule', version.git_version())
-
-    diffoutput = git.diff(R=reverse, M=True, cached=cached,
-                          *argv, **_common_diff_opts())
-    # Handle 'git init'
-    if diffoutput.startswith('fatal:'):
+    status, out, err = git.diff(R=reverse, M=True, cached=cached,
+                                _encoding=encoding,
+                                *argv,
+                                **common_diff_opts())
+    if status != 0:
+        # git init
         if with_diff_header:
             return ('', '')
         else:
             return ''
 
+    return extract_diff_header(status, deleted,
+                               with_diff_header, suppress_header, out)
+
+
+def extract_diff_header(status, deleted,
+                        with_diff_header, suppress_header, diffoutput):
+    headers = []
+
     if diffoutput.startswith('Submodule'):
         if with_diff_header:
             return ('', diffoutput)
         else:
             return diffoutput
 
+    start = False
+    del_tag = 'deleted file mode '
     output = StringIO()
 
     diff = diffoutput.split('\n')
-    for line in map(core.decode, diff):
+    for line in diff:
         if not start and '@@' == line[:2] and '@@' in line[2:]:
             start = True
         if start or (deleted and del_tag in line):
-            output.write(core.encode(line) + '\n')
+            output.write(line + '\n')
         else:
             if with_diff_header:
-                headers.append(core.encode(line))
+                headers.append(line)
             elif not suppress_header:
-                output.write(core.encode(line) + '\n')
+                output.write(line + '\n')
 
-    result = core.decode(output.getvalue())
+    result = output.getvalue()
     output.close()
 
     if with_diff_header:
@@ -293,7 +369,8 @@ def format_patchsets(to_export, revs, output='patches'):
 
     """
 
-    outlines = []
+    outs = []
+    errs = []
 
     cur_rev = to_export[0]
     cur_master_idx = revs.index(cur_rev)
@@ -318,50 +395,48 @@ def format_patchsets(to_export, revs, output='patches'):
     # Export each patchsets
     status = 0
     for patchset in patches_to_export:
-        newstatus, out = export_patchset(patchset[0],
+        stat, out, err = export_patchset(patchset[0],
                                          patchset[-1],
                                          output='patches',
                                          n=len(patchset) > 1,
                                          thread=True,
                                          patch_with_stat=True)
-        outlines.append(out)
-        if status == 0:
-            status += newstatus
-    return (status, '\n'.join(outlines))
+        outs.append(out)
+        if err:
+            errs.append(err)
+        status = max(stat, status)
+    return (status, '\n'.join(outs), '\n'.join(errs))
 
 
 def export_patchset(start, end, output='patches', **kwargs):
     """Export patches from start^ to end."""
-    return git.format_patch('-o', output, start + '^..' + end,
-                            with_stderr=True,
-                            with_status=True,
-                            **kwargs)
+    return git.format_patch('-o', output, start + '^..' + end, **kwargs)
 
 
-def unstage_paths(args):
-    status, output = git.reset('--', with_stderr=True, with_status=True,
-                               *set(args))
-    if status != 128:
-        return (status, output)
-    # handle git init: we have to use 'git rm --cached'
-    # detect this condition by checking if the file is still staged
-    status, output = git.update_index('--',
-                                      force_remove=True,
-                                      with_status=True,
-                                      with_stderr=True,
-                                      *set(args))
-    return (status, output)
+def unstage_paths(args, head='HEAD'):
+    status, out, err = git.reset(head, '--', *set(args))
+    if status == 128:
+        # handle git init: we have to use 'git rm --cached'
+        # detect this condition by checking if the file is still staged
+        return untrack_paths(args, head=head)
+    else:
+        return (status, out, err)
 
 
+def untrack_paths(args, head='HEAD'):
+    if not args:
+        return (-1, N_('Nothing to do'), '')
+    return git.update_index('--', force_remove=True, *set(args))
 
-def worktree_state(head='HEAD', staged_only=False):
+
+def worktree_state(head='HEAD'):
     """Return a tuple of files in various states of being
 
     Can be staged, unstaged, untracked, unmerged, or changed
     upstream.
 
     """
-    state = worktree_state_dict(head=head, staged_only=staged_only)
+    state = worktree_state_dict(head=head)
     return(state.get('staged', []),
            state.get('modified', []),
            state.get('unmerged', []),
@@ -369,105 +444,39 @@ def worktree_state(head='HEAD', staged_only=False):
            state.get('upstream_changed', []))
 
 
-def worktree_state_dict(head='HEAD', staged_only=False, git=git):
+def worktree_state_dict(head='HEAD',
+                        update_index=False,
+                        display_untracked=True,
+                        paths=None):
     """Return a dict of files in various states of being
 
     :rtype: dict, keys are staged, unstaged, untracked, unmerged,
             changed_upstream, and submodule.
 
     """
-    git.update_index(refresh=True)
+    if update_index:
+        git.update_index(refresh=True)
 
-    if staged_only:
-        return _branch_status(head)
+    staged, unmerged, staged_submods = diff_index(head, paths=paths)
+    modified, modified_submods = diff_worktree(paths)
+    untracked = display_untracked and untracked_files(paths=paths) or []
 
-    staged_set = set()
-    modified_set = set()
+    # Remove unmerged paths from the modified list
+    unmerged_set = set(unmerged)
+    modified_set = set(modified)
+    modified_unmerged = modified_set.intersection(unmerged_set)
+    for path in modified_unmerged:
+        modified.remove(path)
 
-    staged = []
-    modified = []
-    unmerged = []
-    untracked = []
-    upstream_changed = []
-    submodules = set()
-    try:
-        output = git.diff_index(head, cached=True, with_stderr=True)
-        if output.startswith('fatal:'):
-            raise errors.GitInitError('git init')
-        for line in output.splitlines():
-            rest, name = line.split('\t', 1)
-            status = rest[-1]
-            name = eval_path(name)
-            if '160000' in rest[1:14]:
-                submodules.add(name)
-            if status  == 'M':
-                staged.append(name)
-                staged_set.add(name)
-                # This file will also show up as 'M' without --cached
-                # so by default don't consider it modified unless
-                # it's truly modified
-                modified_set.add(name)
-                if not staged_only and is_modified(name):
-                    modified.append(name)
-            elif status == 'A':
-                staged.append(name)
-                staged_set.add(name)
-            elif status == 'D':
-                staged.append(name)
-                staged_set.add(name)
-                modified_set.add(name)
-            elif status == 'U':
-                unmerged.append(name)
-                modified_set.add(name)
+    # All submodules
+    submodules = staged_submods.union(modified_submods)
 
-    except errors.GitInitError:
-        # handle git init
-        staged.extend(all_files())
-
-    try:
-        output = git.diff_index(head, with_stderr=True)
-        if output.startswith('fatal:'):
-            raise errors.GitInitError('git init')
-        for line in output.splitlines():
-            rest , name = line.split('\t', 1)
-            status = rest[-1]
-            name = eval_path(name)
-            if '160000' in rest[1:13]:
-                submodules.add(name)
-            if status == 'M' or status == 'D':
-                if name not in modified_set:
-                    modified.append(name)
-            elif status == 'A':
-                # newly-added yet modified
-                if (name not in modified_set and not staged_only and
-                        is_modified(name)):
-                    modified.append(name)
-
-    except errors.GitInitError:
-        # handle git init
-        ls_files = git.ls_files(modified=True, z=True)[:-1].split('\0')
-        modified.extend(map(core.decode, [f for f in ls_files if f]))
-
-    untracked.extend(untracked_files())
+    # Add submodules to the staged and unstaged lists
+    staged.extend(staged_submods)
+    modified.extend(modified_submods)
 
     # Look for upstream modified files if this is a tracking branch
-    tracked = tracked_branch()
-    if tracked:
-        try:
-            diff_expr = merge_base_to(tracked)
-            output = git.diff(diff_expr, name_only=True, z=True,
-                              **_common_diff_opts())
-
-            if output.startswith('fatal:'):
-                raise errors.GitInitError('git init')
-
-            for name in [n for n in output.split('\0') if n]:
-                name = core.decode(name)
-                upstream_changed.append(name)
-
-        except errors.GitInitError:
-            # handle git init
-            pass
+    upstream_changed = diff_upstream(head)
 
     # Keep stuff sorted
     staged.sort()
@@ -484,75 +493,98 @@ def worktree_state_dict(head='HEAD', staged_only=False, git=git):
             'submodules': submodules}
 
 
-def _branch_status(branch, git=git):
+def diff_index(head, cached=True, paths=None):
+    submodules = set()
+    staged = []
+    unmerged = []
+
+    if paths is None:
+        paths = []
+    filter_paths = [head, '--'] + paths
+    status, out, err = git.diff_index(cached=cached, z=True,
+                                      *filter_paths)
+    if status != 0:
+        # handle git init
+        return tracked_files(), unmerged, submodules
+
+    while out:
+        rest, out = out.split('\0', 1)
+        name, out = out.split('\0', 1)
+        status = rest[-1]
+        if '160000' in rest[1:14]:
+            submodules.add(name)
+        elif status  in 'DAMT':
+            staged.append(name)
+        elif status == 'U':
+            unmerged.append(name)
+
+    return staged, unmerged, submodules
+
+
+def diff_worktree(paths=None):
+    modified = []
+    submodules = set()
+
+    if paths is None:
+        paths = []
+    filter_paths = ['--'] + paths
+    status, out, err = git.diff_files(z=True, *filter_paths)
+    if status != 0:
+        # handle git init
+        out = git.ls_files(modified=True, z=True)[STDOUT]
+        if out:
+            modified = out[:-1].split('\0')
+        return modified, submodules
+
+    while out:
+        rest, out = out.split('\0', 1)
+        name, out = out.split('\0', 1)
+        status = rest[-1]
+        if '160000' in rest[1:14]:
+            submodules.add(name)
+        elif status in 'DAMT':
+            modified.append(name)
+
+    return modified, submodules
+
+
+def diff_upstream(head):
+    tracked = tracked_branch()
+    if not tracked:
+        return []
+    base = merge_base(head, tracked)
+    return diff_filenames(base, tracked)
+
+
+def _branch_status(branch):
     """
     Returns a tuple of staged, unstaged, untracked, and unmerged files
 
     This shows only the changes that were introduced in branch
 
     """
-    status, output = git.diff(name_only=True,
-                              M=True, z=True,
-                              with_status=True,
-                              *branch.strip().split(),
-                              **_common_diff_opts())
-    if status != 0:
-        return {}
-
-    staged = map(core.decode, [n for n in output.split('\0') if n])
+    staged = diff_filenames(branch)
     return {'staged': staged,
             'upstream_changed': staged}
 
 
-def merge_base_to(ref):
+def merge_base(head, ref):
     """Given `ref`, return $(git merge-base ref HEAD)..ref."""
-    base = git.merge_base('HEAD', ref)
-    return '%s..%s' % (base, ref)
+    return git.merge_base(head, ref)[STDOUT]
 
 
 def merge_base_parent(branch):
     tracked = tracked_branch(branch=branch)
     if tracked:
-        return '%s..%s' % (tracked, branch)
-    return 'master..%s' % branch
-
-
-def is_modified(name, git=git):
-    status, out = git.diff('--', name,
-                           name_only=True,
-                           exit_code=True,
-                           with_status=True,
-                           **_common_diff_opts())
-    return status != 0
-
-
-def eval_path(path):
-    """handles quoted paths."""
-    if path.startswith('"') and path.endswith('"'):
-        return core.decode(eval(path))
-    else:
-        return core.decode(path)
-
-
-def renamed_files(start, end, git=git):
-    difflines = git.diff('%s..%s' % (start, end), M=True,
-                         **_common_diff_opts()).splitlines()
-    return [eval_path(r[12:].rstrip())
-                for r in difflines if r.startswith('rename from ')]
-
-
-def changed_files(start, end, git=git):
-    zfiles_str = git.diff('%s..%s' % (start, end),
-                          name_only=True, z=True,
-                          **_common_diff_opts()).strip('\0')
-    return [core.decode(enc) for enc in zfiles_str.split('\0') if enc]
+        return tracked
+    return 'HEAD'
 
 
 def parse_ls_tree(rev):
     """Return a list of(mode, type, sha1, path) tuples."""
-    lines = git.ls_tree(rev, r=True).splitlines()
     output = []
-    regex = re.compile('^(\d+)\W(\w+)\W(\w+)[ \t]+(.*)$')
+    lines = git.ls_tree(rev, r=True)[STDOUT].splitlines()
+    regex = re.compile(r'^(\d+)\W(\w+)\W(\w+)[ \t]+(.*)$')
     for line in lines:
         match = regex.match(line)
         if match:
@@ -565,12 +597,12 @@ def parse_ls_tree(rev):
 
 
 # A regex for matching the output of git(log|rev-list) --pretty=oneline
-REV_LIST_REGEX = re.compile('^([0-9a-f]{40}) (.*)$')
+REV_LIST_REGEX = re.compile(r'^([0-9a-f]{40}) (.*)$')
 
 def parse_rev_list(raw_revs):
     """Parse `git log --pretty=online` output into (SHA-1, summary) pairs."""
     revs = []
-    for line in map(core.decode, raw_revs.splitlines()):
+    for line in raw_revs.splitlines():
         match = REV_LIST_REGEX.match(line)
         if match:
             rev_id = match.group(1)
@@ -586,8 +618,8 @@ def log_helper(all=False, extra_args=None):
     args = []
     if extra_args:
         args = extra_args
-    output = git.log(pretty='oneline', no_color=True, all=all, *args)
-    for line in map(core.decode, output.splitlines()):
+    output = log(git, pretty='oneline', all=all, *args)
+    for line in output.splitlines():
         match = REV_LIST_REGEX.match(line)
         if match:
             revs.append(match.group(1))
@@ -598,15 +630,23 @@ def log_helper(all=False, extra_args=None):
 def rev_list_range(start, end):
     """Return a (SHA-1, summary) pairs between start and end."""
     revrange = '%s..%s' % (start, end)
-    raw_revs = git.rev_list(revrange, pretty='oneline')
-    return parse_rev_list(raw_revs)
+    out = git.rev_list(revrange, pretty='oneline')[STDOUT]
+    return parse_rev_list(out)
+
+
+def commit_message_path():
+    """Return the path to .git/GIT_COLA_MSG"""
+    path = git.git_path("GIT_COLA_MSG")
+    if core.exists(path):
+        return path
+    return None
 
 
 def merge_message_path():
     """Return the path to .git/MERGE_MSG or .git/SQUASH_MSG."""
     for basename in ('MERGE_MSG', 'SQUASH_MSG'):
         path = git.git_path(basename)
-        if os.path.exists(path):
+        if core.exists(path):
             return path
     return None
 
@@ -617,18 +657,18 @@ def abort_merge():
     git.read_tree('HEAD', reset=True, u=True, v=True)
     # remove MERGE_HEAD
     merge_head = git.git_path('MERGE_HEAD')
-    if os.path.exists(merge_head):
-        os.unlink(merge_head)
+    if core.exists(merge_head):
+        core.unlink(merge_head)
     # remove MERGE_MESSAGE, etc.
     merge_msg_path = merge_message_path()
     while merge_msg_path:
-        os.unlink(merge_msg_path)
+        core.unlink(merge_msg_path)
         merge_msg_path = merge_message_path()
 
 
 def merge_message(revision):
     """Return a merge message for FETCH_HEAD."""
     fetch_head = git.git_path('FETCH_HEAD')
-    if os.path.exists(fetch_head):
-        return git.fmt_merge_msg('--file', fetch_head)
+    if core.exists(fetch_head):
+        return git.fmt_merge_msg('--file', fetch_head)[STDOUT]
     return "Merge branch '%s'" % revision
