@@ -1,22 +1,26 @@
+from __future__ import division, absolute_import, unicode_literals
+
 import os
 import re
 
+from PyQt4 import QtCore
 from PyQt4 import QtGui
+from PyQt4.QtCore import SIGNAL
 
-import cola
 from cola import cmds
 from cola import core
 from cola import difftool
 from cola import gitcmds
 from cola import qtutils
+from cola import utils
 from cola.git import git
 from cola.i18n import N_
 from cola.interaction import Interaction
+from cola.models import main
 from cola.widgets import completion
 from cola.widgets.browse import BrowseDialog
-from cola.widgets.grep import run_grep
 from cola.widgets.selectcommits import select_commits
-
+from cola.compat import ustr
 
 
 def delete_branch():
@@ -39,7 +43,6 @@ def delete_remote_branch():
         branch = match.group('branch')
         cmds.do(cmds.DeleteRemoteBranch, remote, branch)
 
-
 def browse_current():
     """Launch the 'Browse Current Branch' dialog."""
     branch = gitcmds.current_branch()
@@ -89,7 +92,7 @@ def new_repo():
     paths = dlg.selectedFiles()
     if not paths:
         return None
-    path = unicode(paths[0])
+    path = ustr(paths[0])
     if not path:
         return None
     # Avoid needlessly calling `git init`.
@@ -121,15 +124,15 @@ def open_new_repo():
     cmds.do(cmds.OpenRepo, dirname)
 
 
-def clone_repo(spawn=True):
+def prompt_for_clone():
     """
-    Present GUI controls for cloning a repository
+    Present a GUI for cloning a repository.
 
-    A new cola session is invoked when 'spawn' is True.
+    Returns the target directory and URL
 
     """
     url, ok = qtutils.prompt(N_('Path or URL to clone (Env. $VARS okay)'))
-    url = os.path.expandvars(url)
+    url = utils.expandpath(url)
     if not ok or not url:
         return None
     try:
@@ -156,7 +159,7 @@ def clone_repo(spawn=True):
 
     # Prompt the user for a directory to use as the parent directory
     msg = N_('Select a parent directory for the new clone')
-    dirname = qtutils.opendir_dialog(msg, cola.model().getcwd())
+    dirname = qtutils.opendir_dialog(msg, main.model().getcwd())
     if not dirname:
         return None
     count = 1
@@ -166,15 +169,14 @@ def clone_repo(spawn=True):
         # An existing path can be specified
         msg = (N_('"%s" already exists, cola will create a new directory') %
                destdir)
-        Interaction.information('Directory Exists', msg)
+        Interaction.information(N_('Directory Exists'), msg)
 
     # Make sure the new destdir doesn't exist
     while core.exists(destdir):
         destdir = olddestdir + str(count)
         count += 1
-    if cmds.do(cmds.Clone, url, destdir, spawn=spawn):
-        return destdir
-    return None
+
+    return url, destdir
 
 
 def export_patches():
@@ -183,9 +185,7 @@ def export_patches():
     to_export = select_commits(N_('Export Patches'), revs, summaries)
     if not to_export:
         return
-    to_export.reverse()
-    revs.reverse()
-    cmds.do(cmds.FormatPatch, to_export, revs)
+    cmds.do(cmds.FormatPatch, reversed(to_export), reversed(revs))
 
 
 def diff_expression():
@@ -199,27 +199,27 @@ def diff_expression():
     difftool.diff_expression(qtutils.active_window(), ref)
 
 
-def grep():
-    """Prompt and use 'git grep' to find the content."""
-    widget = run_grep(parent=qtutils.active_window())
-    widget.show()
-    widget.raise_()
-    return widget
-
-
 def open_repo():
-    """Spawn a new cola session."""
     dirname = qtutils.opendir_dialog(N_('Open Git Repository...'),
-                                     cola.model().getcwd())
+                                     main.model().getcwd())
     if not dirname:
         return
     cmds.do(cmds.OpenRepo, dirname)
 
 
+def open_repo_in_new_window():
+    """Spawn a new cola session."""
+    dirname = qtutils.opendir_dialog(N_('Open Git Repository...'),
+                                     main.model().getcwd())
+    if not dirname:
+        return
+    cmds.do(cmds.OpenNewRepo, dirname)
+
+
 def load_commitmsg():
     """Load a commit message from a file."""
-    filename = qtutils.open_dialog(N_('Load Commit Message'),
-                                   cola.model().getcwd())
+    filename = qtutils.open_file(N_('Load Commit Message'),
+                                 directory=main.model().getcwd())
     if filename:
         cmds.do(cmds.LoadCommitMessageFromFile, filename)
 
@@ -251,3 +251,107 @@ def review_branch():
         return
     merge_base = gitcmds.merge_base_parent(branch)
     difftool.diff_commits(qtutils.active_window(), merge_base, branch)
+
+
+class TaskRunner(QtCore.QObject):
+    """Runs QRunnable instances and transfers control when they finish"""
+
+    def __init__(self, parent):
+        QtCore.QObject.__init__(self, parent)
+        self.tasks = []
+        self.task_details = {}
+        self.connect(self, Task.FINISHED, self.finish)
+
+    def start(self, task, progress=None, finish=None):
+        """Start the task and register a callback"""
+        if progress is not None:
+            progress.show()
+        # prevents garbage collection bugs in certain PyQt4 versions
+        self.tasks.append(task)
+        task_id = id(task)
+        self.task_details[task_id] = (progress, finish)
+        QtCore.QThreadPool.globalInstance().start(task)
+
+    def finish(self, task, *args, **kwargs):
+        task_id = id(task)
+        try:
+            self.tasks.remove(task)
+        except:
+            pass
+        try:
+            progress, finish = self.task_details[task_id]
+            del self.task_details[task_id]
+        except KeyError:
+            finish = progress = None
+
+        if progress is not None:
+            progress.hide()
+
+        if finish is not None:
+            finish(task, *args, **kwargs)
+
+
+class Task(QtCore.QRunnable):
+    """Base class for concrete tasks"""
+
+    FINISHED = SIGNAL('finished')
+
+    def __init__(self, sender):
+        QtCore.QRunnable.__init__(self)
+        self.sender = sender
+
+    def finish(self, *args, **kwargs):
+        self.sender.emit(self.FINISHED, self, *args, **kwargs)
+
+
+class CloneTask(Task):
+    """Clones a Git repository"""
+
+    def __init__(self, sender, url, destdir, spawn):
+        Task.__init__(self, sender)
+        self.url = url
+        self.destdir = destdir
+        self.spawn = spawn
+        self.cmd = None
+
+    def run(self):
+        """Runs the model action and captures the result"""
+        self.cmd = cmds.do(cmds.Clone, self.url, self.destdir,
+                           spawn=self.spawn)
+        self.finish()
+
+
+def clone_repo(task_runner, progress, finish, spawn):
+    """Clone a repostiory asynchronously with progress animation"""
+    result = prompt_for_clone()
+    if result is None:
+        return
+    # Use a thread to update in the background
+    url, destdir = result
+    progress.set_details(N_('Clone Repository'),
+                         N_('Cloning repository at %s') % url)
+    task = CloneTask(task_runner, url, destdir, spawn)
+    task_runner.start(task,
+                      finish=finish,
+                      progress=progress)
+
+
+def report_clone_repo_errors(task):
+    """Report errors from the clone task if they exist"""
+    if task.cmd is None or task.cmd.ok:
+        return
+    Interaction.critical(task.cmd.error_message,
+                         message=task.cmd.error_message,
+                         details=task.cmd.error_details)
+
+def rename_branch():
+    """Launch the 'Rename Branch' dialogs."""
+    branch = choose_branch(N_('Rename Existing Branch'), N_('Select'))
+    if not branch:
+        return
+    new_branch = choose_branch(N_('Enter Branch New Name'), N_('Rename'))
+    if not new_branch:
+        return
+    cmds.do(cmds.RenameBranch, branch, new_branch)
+
+
