@@ -1,14 +1,16 @@
+from __future__ import division, absolute_import, unicode_literals
+
 from PyQt4 import QtGui
 from PyQt4 import QtCore
 from PyQt4.QtCore import Qt
 from PyQt4.QtCore import SIGNAL
 
-import cola
 from cola import cmds
 from cola import core
 from cola import gitcmds
 from cola import gitcfg
 from cola import textwrap
+from cola import qtutils
 from cola.cmds import Interaction
 from cola.gitcmds import commit_message_path
 from cola.i18n import N_
@@ -23,12 +25,14 @@ from cola.qtutils import connect_action_bool
 from cola.qtutils import connect_button
 from cola.qtutils import create_toolbutton
 from cola.qtutils import diff_font
+from cola.qtutils import hide_button_menu_indicator
 from cola.qtutils import options_icon
 from cola.qtutils import save_icon
 from cola.widgets import defs
 from cola.widgets.selectcommits import select_commits
 from cola.widgets.spellcheck import SpellCheckTextEdit
 from cola.widgets.text import HintedLineEdit
+from cola.compat import ustr
 
 
 class CommitMessageEditor(QtGui.QWidget):
@@ -57,6 +61,7 @@ class CommitMessageEditor(QtGui.QWidget):
 
         # Widgets
         self.summary = CommitSummaryLineEdit()
+        self.summary.setMinimumHeight(defs.tool_button_height)
         self.summary.extra_actions.append(self.signoff_action)
         self.summary.extra_actions.append(self.commit_action)
 
@@ -75,6 +80,7 @@ class CommitMessageEditor(QtGui.QWidget):
                                                 tooltip=N_('Actions...'))
         self.actions_button.setMenu(self.actions_menu)
         self.actions_button.setPopupMode(QtGui.QToolButton.InstantPopup)
+        hide_button_menu_indicator(self.actions_button)
 
         self.actions_menu.addAction(self.signoff_action)
         self.actions_menu.addAction(self.commit_action)
@@ -87,6 +93,13 @@ class CommitMessageEditor(QtGui.QWidget):
         self.amend_action.setShortcut(cmds.AmendMode.SHORTCUT)
         self.amend_action.setShortcutContext(Qt.ApplicationShortcut)
 
+        # Sign commits
+        cfg = gitcfg.current()
+        self.sign_action = self.actions_menu.addAction(
+                N_('Create Signed Commit'))
+        self.sign_action.setCheckable(True)
+        self.sign_action.setChecked(cfg.get('cola.signcommits', False))
+
         # Spell checker
         self.check_spelling_action = self.actions_menu.addAction(
                 N_('Check Spelling'))
@@ -111,25 +124,18 @@ class CommitMessageEditor(QtGui.QWidget):
         self.connect(self.fixup_commit_menu, SIGNAL('aboutToShow()'),
                      self.build_fixup_menu)
 
-        self.toplayout = QtGui.QHBoxLayout()
-        self.toplayout.setMargin(0)
-        self.toplayout.setSpacing(defs.spacing)
-        self.toplayout.addWidget(self.actions_button)
-        self.toplayout.addWidget(self.summary)
-        self.toplayout.addWidget(self.commit_button)
+        self.toplayout = qtutils.hbox(defs.no_margin, defs.spacing,
+                                      self.actions_button, self.summary,
+                                      self.commit_button)
+        self.toplayout.setContentsMargins(defs.margin, defs.no_margin,
+                                          defs.no_margin, defs.no_margin)
 
-        self.mainlayout = QtGui.QVBoxLayout()
-        self.mainlayout.setMargin(defs.margin)
-        self.mainlayout.setSpacing(defs.spacing)
-        self.mainlayout.addLayout(self.toplayout)
-        self.mainlayout.addWidget(self.description)
+        self.mainlayout = qtutils.vbox(defs.no_margin, defs.spacing,
+                                       self.toplayout, self.description)
         self.setLayout(self.mainlayout)
 
         connect_button(self.commit_button, self.commit)
 
-        notifier = cola.notifier()
-        notifier.connect(notifier.AMEND, self.amend_action.setChecked)
-
         # Broadcast the amend mode
         connect_action_bool(self.amend_action, cmds.run(cmds.AmendMode))
         connect_action_bool(self.check_spelling_action,
@@ -139,7 +145,10 @@ class CommitMessageEditor(QtGui.QWidget):
         connect_action_bool(self.autowrap_action, self.set_linebreak)
 
         add_action(self.summary, N_('Move Down'), self.focus_description,
-                Qt.Key_Down, Qt.Key_Return, Qt.Key_Enter)
+                   Qt.Key_Return, Qt.Key_Enter)
+
+        add_action(self.summary, N_('Move Down'),
+                   self.summary_cursor_down, Qt.Key_Down)
 
         self.model.add_observer(self.model.message_commit_message_changed,
                                 self.set_commit_message)
@@ -153,7 +162,7 @@ class CommitMessageEditor(QtGui.QWidget):
 
         # Keep model informed of changes
         self.connect(self.summary, SIGNAL('textChanged(QString)'),
-                     self.commit_message_changed)
+                     self.commit_summary_changed)
 
         self.connect(self.description, SIGNAL('textChanged()'),
                      self.commit_message_changed)
@@ -198,6 +207,21 @@ class CommitMessageEditor(QtGui.QWidget):
     def focus_description(self):
         self.description.setFocus()
 
+    def summary_cursor_down(self):
+        """Handle the down key in the summary field
+
+        If the cursor is at the end of the line then focus the description.
+        Otherwise, move the cursor to the end of the line so that a
+        subsequence "down" press moves to the end of the line.
+
+        """
+        cur_position = self.summary.cursorPosition()
+        end_position = len(self.summary.value())
+        if cur_position == end_position:
+            self.focus_description()
+        else:
+            self.summary.setCursorPosition(end_position)
+
     def commit_message(self, raw=True):
         """Return the commit message as a unicode string"""
         summary = self.summary.value()
@@ -220,6 +244,30 @@ class CommitMessageEditor(QtGui.QWidget):
             return text
         return textwrap.word_wrap(text, self._tabwidth, self._textwidth)
 
+    def commit_summary_changed(self, value):
+        """Respond to changes to the `summary` field
+
+        Newlines can enter the `summary` field when pasting, which is
+        undesirable.  Break the pasted value apart into the separate
+        (summary, description) values and move the description over to the
+        "extended description" field.
+
+        """
+        value = ustr(value)
+        if '\n' in value:
+            summary, description = value.split('\n', 1)
+            description = description.lstrip('\n')
+            cur_description = self.description.value()
+            if cur_description:
+                description = description + '\n' + cur_description
+            # this callback is triggered by changing `summary`
+            # so disable signals for `summary` only.
+            self.summary.blockSignals(True)
+            self.summary.set_value(summary)
+            self.summary.blockSignals(False)
+            self.description.set_value(description)
+        self.commit_message_changed()
+
     def commit_message_changed(self, value=None):
         """Update the model when values change"""
         self.notifying = True
@@ -247,7 +295,7 @@ class CommitMessageEditor(QtGui.QWidget):
             return
 
         # Parse the "summary" and "description" fields
-        umsg = unicode(message)
+        umsg = ustr(message)
         lines = umsg.splitlines()
 
         num_lines = len(lines)
@@ -333,8 +381,10 @@ class CommitMessageEditor(QtGui.QWidget):
         self.description.setFont(font)
 
     def set_mode(self, mode):
+        can_amend = not self.model.is_merging
         checked = (mode == self.model.mode_amend)
         blocksignals = self.amend_action.blockSignals(True)
+        self.amend_action.setEnabled(can_amend)
         self.amend_action.setChecked(checked)
         self.amend_action.blockSignals(blocksignals)
 
@@ -387,7 +437,8 @@ class CommitMessageEditor(QtGui.QWidget):
                         N_('Amend Commit'),
                         default=False, icon=save_icon())):
             return
-        status, out, err = cmds.do(cmds.Commit, amend, msg)
+        sign = self.sign_action.isChecked()
+        status, out, err = cmds.do(cmds.Commit, amend, msg, sign)
         if status != 0:
             Interaction.critical(N_('Commit failed'),
                                  N_('"git commit" returned exit code %s') %
@@ -445,7 +496,7 @@ class CommitMessageEditor(QtGui.QWidget):
         if enabled and not self.spellcheck_initialized:
             # Add our name to the dictionary
             self.spellcheck_initialized = True
-            cfg = gitcfg.instance()
+            cfg = gitcfg.current()
             user_name = cfg.get('user.name')
             if user_name:
                 for part in user_name.split():
@@ -483,28 +534,15 @@ class CommitSummaryLineEdit(HintedLineEdit):
 
 
 class CommitMessageTextEdit(SpellCheckTextEdit):
+
     def __init__(self, parent=None):
         hint = N_('Extended description...')
         SpellCheckTextEdit.__init__(self, hint, parent)
         self.extra_actions = []
-        self.setMinimumSize(QtCore.QSize(1, 1))
 
         self.action_emit_leave = add_action(self,
                 'Shift Tab', self.emit_leave, 'Shift+tab')
 
-        self.installEventFilter(self)
-
-    def eventFilter(self, obj, event):
-        if event.type() == QtCore.QEvent.FocusIn:
-            height = QtGui.QFontMetrics(self.font()).height() * 3
-            height += defs.spacing * 4
-            self.setMinimumSize(QtCore.QSize(1, height))
-
-        elif event.type() == QtCore.QEvent.FocusOut:
-            self.setMinimumSize(QtCore.QSize(1, 1))
-
-        return False
-
     def contextMenuEvent(self, event):
         menu, spell_menu = self.context_menu()
         if self.extra_actions:
@@ -529,7 +567,7 @@ class CommitMessageTextEdit(SpellCheckTextEdit):
                     self.emit_leave()
                 event.accept()
                 return
-            text_before = unicode(self.toPlainText())[:position]
+            text_before = ustr(self.toPlainText())[:position]
             lines_before = text_before.count('\n')
             if lines_before == 0:
                 # If we're on the first line, but not at the
@@ -546,7 +584,7 @@ class CommitMessageTextEdit(SpellCheckTextEdit):
         elif event.key() == Qt.Key_Down:
             cursor = self.textCursor()
             position = cursor.position()
-            all_text = unicode(self.toPlainText())
+            all_text = ustr(self.toPlainText())
             text_after = all_text[position:]
             lines_after = text_after.count('\n')
             if lines_after == 0:
@@ -562,3 +600,8 @@ class CommitMessageTextEdit(SpellCheckTextEdit):
 
     def emit_leave(self):
         self.emit(SIGNAL('leave()'))
+
+    def setFont(self, font):
+        SpellCheckTextEdit.setFont(self, font)
+        fm = self.fontMetrics()
+        self.setMinimumSize(QtCore.QSize(1, fm.height() * 2))
