@@ -1,3 +1,5 @@
+from __future__ import division, absolute_import, unicode_literals
+
 import re
 import subprocess
 
@@ -6,25 +8,25 @@ from PyQt4 import QtGui
 from PyQt4.QtCore import Qt
 from PyQt4.QtCore import SIGNAL
 
-import cola
+from cola import gitcmds
 from cola.i18n import N_
 from cola import qtutils
 from cola import utils
-from cola.compat import set
+from cola.models import main
 from cola.widgets import defs
+from cola.widgets import text
+from cola.compat import ustr
 
 
-class CompletionLineEdit(QtGui.QLineEdit):
+UPDATE_SIGNAL = 'update()'
 
-    def __init__(self, model, parent=None):
-        QtGui.QLineEdit.__init__(self, parent)
+
+class CompletionLineEdit(text.HintedLineEdit):
+
+    def __init__(self, model, hint='', parent=None):
+        text.HintedLineEdit.__init__(self, hint=hint, parent=parent)
 
         self.setFont(qtutils.diff_font())
-        # used to hide the completion popup after a drag-select
-        self._drag = 0
-
-        self._keys_to_ignore = set([Qt.Key_Enter, Qt.Key_Return,
-                                    Qt.Key_Escape])
 
         completion_model = model(self)
         completer = Completer(completion_model, self)
@@ -39,6 +41,14 @@ class CompletionLineEdit(QtGui.QLineEdit):
         self.connect(self._completer, SIGNAL('activated(QString)'),
                      self._complete)
 
+        self.connect(self, SIGNAL('destroyed(QObject*)'), self.dispose)
+
+    def __del__(self):
+        self.dispose()
+
+    def dispose(self, *args):
+        self._completer.dispose()
+
     def refresh(self):
         return self._completer.model().update()
 
@@ -46,23 +56,24 @@ class CompletionLineEdit(QtGui.QLineEdit):
         return self._completer.popup()
 
     def value(self):
-        return unicode(self.text())
+        return ustr(self.text())
 
     def _is_case_sensitive(self, text):
         return bool([char for char in text if char.isupper()])
 
     def _text_changed(self, text):
-        text = self._last_word()
-        self._do_text_changed(text)
+        match_text = self._last_word()
+        full_text = ustr(text)
+        self._do_text_changed(full_text, match_text)
 
-    def _do_text_changed(self, text):
-        case_sensitive = self._is_case_sensitive(text)
+    def _do_text_changed(self, full_text, match_text):
+        case_sensitive = self._is_case_sensitive(match_text)
         if case_sensitive:
             self._completer.setCaseSensitivity(Qt.CaseSensitive)
         else:
             self._completer.setCaseSensitivity(Qt.CaseInsensitive)
-        self._delegate.set_highlight_text(text, case_sensitive)
-        self._completer.set_match_text(text, case_sensitive)
+        self._delegate.set_highlight_text(match_text, case_sensitive)
+        self._completer.set_match_text(full_text, match_text, case_sensitive)
 
     def update_matches(self):
         text = self._last_word()
@@ -74,9 +85,9 @@ class CompletionLineEdit(QtGui.QLineEdit):
         This is the event handler for the QCompleter.activated(QString) signal,
         it is called when the user selects an item in the completer popup.
         """
-        completion = unicode(completion)
+        completion = ustr(completion)
         if not completion:
-            self._do_text_changed('')
+            self._do_text_changed('', '')
             return
         words = self._words()
         if words and not self._ends_with_whitespace():
@@ -86,82 +97,45 @@ class CompletionLineEdit(QtGui.QLineEdit):
         text = subprocess.list2cmdline(words)
         self.setText(text)
         self.emit(SIGNAL('changed()'))
-        self._do_text_changed('')
+        self._do_text_changed(text, '')
 
     def _words(self):
-        return utils.shell_split(unicode(self.text()))
+        return utils.shell_split(self.value())
 
     def _ends_with_whitespace(self):
-        text = unicode(self.text())
-        return text.rstrip() != text
+        return self.value() != self.value().rstrip()
 
     def _last_word(self):
         if self._ends_with_whitespace():
-            return u''
+            return ''
         words = self._words()
         if not words:
-            return unicode(self.text())
+            return self.value()
         if not words[-1]:
-            return u''
+            return ''
         return words[-1]
 
     def event(self, event):
-        if event.type() == QtCore.QEvent.KeyPress:
-            if (event.key() == Qt.Key_Tab and
-                    self.popup().isVisible()):
-                event.ignore()
-                return True
-            if (event.key() in (Qt.Key_Return, Qt.Key_Enter) and
-                    not self.popup().isVisible()):
-                self.emit(SIGNAL('returnPressed()'))
-                event.accept()
-                return True
         if event.type() == QtCore.QEvent.Hide:
             self.close_popup()
-        return QtGui.QLineEdit.event(self, event)
+        return text.HintedLineEdit.event(self, event)
 
     def do_completion(self):
-        self._completer.popup().setCurrentIndex(
-                self._completer.model().index(0,0))
         self._completer.complete()
+        self.popup().setCurrentIndex(self._completer.model().index(0, 0))
+        self.popup().scrollToTop()
 
-    def keyPressEvent(self, event):
-        if self._completer.popup().isVisible():
-            if event.key() in self._keys_to_ignore:
-                event.ignore()
-                self._complete(self._last_word())
-                return
+    def keyReleaseEvent(self, event):
 
-        elif (event.key() == Qt.Key_Down and
-              self._completer.completionCount() > 0):
-                event.accept()
-                self.do_completion()
-                return
+        text.HintedLineEdit.keyReleaseEvent(self, event)
 
-        QtGui.QLineEdit.keyPressEvent(self, event)
+        if event.key() in (Qt.Key_Tab, Qt.Key_Enter, Qt.Key_Return,
+                           Qt.Key_Escape):
+            return
 
         prefix = self._last_word()
-        if prefix != unicode(self._completer.completionPrefix()):
-            self._update_popup_items(prefix)
-
-        if len(event.text()) > 0 and len(prefix) > 0:
-            self._completer.complete()
-
-    #: _drag: 0 - unclicked, 1 - clicked, 2 - dragged
-    def mousePressEvent(self, event):
-        self._drag = 1
-        return QtGui.QLineEdit.mousePressEvent(self, event)
-
-    def mouseMoveEvent(self, event):
-        if self._drag == 1:
-            self._drag = 2
-        return QtGui.QLineEdit.mouseMoveEvent(self, event)
-
-    def mouseReleaseEvent(self, event):
-        if self._drag != 2 and event.button() != Qt.RightButton:
-            self.do_completion()
-        self._drag = 0
-        return QtGui.QLineEdit.mouseReleaseEvent(self, event)
+        self._update_popup_items(prefix)
+        self.do_completion()
 
     def close_popup(self):
         if self.popup().isVisible():
@@ -173,17 +147,10 @@ class CompletionLineEdit(QtGui.QLineEdit):
         with the given prefix.
         """
         self._completer.setCompletionPrefix(prefix)
-        self._completer.popup().setCurrentIndex(
-                self._completer.model().index(0,0))
-
-    def __del__(self):
-        self.dispose()
-
-    def dispose(self):
-        self._completer.dispose()
 
 
 class GatherCompletionsThread(QtCore.QThread):
+
     def __init__(self, model):
         QtCore.QThread.__init__(self)
         self.model = model
@@ -193,9 +160,9 @@ class GatherCompletionsThread(QtCore.QThread):
         text = None
         # Loop when the matched text changes between the start and end time.
         # This happens when gather_matches() takes too long and the
-        # model's matched_text changes in-between.
-        while text != self.model.matched_text:
-            text = self.model.matched_text
+        # model's match_text changes in-between.
+        while text != self.model.match_text:
+            text = self.model.match_text
             items = self.model.gather_matches(self.case_sensitive)
 
         if text is not None:
@@ -225,7 +192,7 @@ class HighlightDelegate(QtGui.QStyledItemDelegate):
         if not self.highlight_text:
             return QtGui.QStyledItemDelegate.paint(self, painter, option, index)
 
-        text = unicode(index.data().toPyObject())
+        text = ustr(index.data().toPyObject())
         if self.case_sensitive:
             html = text.replace(self.highlight_text,
                                 '<strong>%s</strong>' % self.highlight_text)
@@ -272,25 +239,21 @@ class CompletionModel(QtGui.QStandardItemModel):
 
     def __init__(self, parent):
         QtGui.QStandardItemModel.__init__(self, parent)
-        self.matched_text = ''
+        self.match_text = ''
+        self.full_text = ''
         self.case_sensitive = False
 
         self.update_thread = GatherCompletionsThread(self)
         self.connect(self.update_thread, SIGNAL('items_gathered'),
                      self.apply_matches)
 
-    def lower_completion_cmp(self, a, b):
-        return cmp(a.replace('.','').lower(), a.replace('.','').lower())
-
-    def completion_cmp(self, a, b):
-        return cmp(a.replace('.',''), a.replace('.',''))
-
     def update(self):
         case_sensitive = self.update_thread.case_sensitive
         self.update_matches(case_sensitive)
 
-    def set_match_text(self, matched_text, case_sensitive):
-        self.matched_text = matched_text
+    def set_match_text(self, full_text, match_text, case_sensitive):
+        self.full_text = full_text
+        self.match_text = match_text
         self.update_matches(case_sensitive)
 
     def update_matches(self, case_sensitive):
@@ -310,7 +273,6 @@ class CompletionModel(QtGui.QStandardItemModel):
         dir_icon = qtutils.dir_icon()
         git_icon = qtutils.git_icon()
 
-        matched_text = self.matched_text
         items = []
         for ref in matched_refs:
             item = QStandardItem()
@@ -318,12 +280,6 @@ class CompletionModel(QtGui.QStandardItemModel):
             item.setIcon(git_icon)
             items.append(item)
 
-        if matched_paths and (not matched_text or matched_text in '--'):
-            item = QStandardItem()
-            item.setText('--')
-            item.setIcon(file_icon)
-            items.append(item)
-
         for match in matched_paths:
             item = QStandardItem()
             item.setText(match)
@@ -337,6 +293,37 @@ class CompletionModel(QtGui.QStandardItemModel):
         self.invisibleRootItem().appendRows(items)
 
 
+def filter_matches(match_text, candidates, case_sensitive):
+    """Filter candidates and return the matches"""
+
+    if case_sensitive:
+        transform = lambda x: x
+        keyfunc = lambda x: x.replace('.', '')
+    else:
+        transform = lambda x: x.lower()
+        keyfunc = lambda x: x.replace('.', '').lower()
+
+    if match_text:
+        matches = [r for r in candidates
+                    if transform(match_text) in transform(r)]
+    else:
+        matches = list(candidates)
+
+    matches.sort(key=keyfunc)
+    return matches
+
+
+def filter_path_matches(match_text, file_list, case_sensitive):
+    """Return matching completions from a list of candidate files"""
+
+    files = set(file_list)
+    files_and_dirs = utils.add_parents(files)
+    dirs = files_and_dirs.difference(files)
+
+    paths = filter_matches(match_text, files_and_dirs, case_sensitive)
+    return (paths, dirs)
+
+
 class Completer(QtGui.QCompleter):
 
     def __init__(self, model, parent):
@@ -345,7 +332,7 @@ class Completer(QtGui.QCompleter):
         self.setCompletionMode(QtGui.QCompleter.UnfilteredPopupCompletion)
         self.setCaseSensitivity(Qt.CaseInsensitive)
 
-        self.connect(model, SIGNAL('update()'), self.update)
+        self.connect(model, SIGNAL(UPDATE_SIGNAL), self.update)
         self.setModel(model)
 
     def update(self):
@@ -354,47 +341,33 @@ class Completer(QtGui.QCompleter):
     def dispose(self):
         self._model.dispose()
 
-    def set_match_text(self, matched_text, case_sensitive):
-        self._model.set_match_text(matched_text, case_sensitive)
+    def set_match_text(self, full_text, match_text, case_sensitive):
+        self._model.set_match_text(full_text, match_text, case_sensitive)
 
 
 class GitCompletionModel(CompletionModel):
 
     def __init__(self, parent):
         CompletionModel.__init__(self, parent)
-        self.cola_model = model = cola.model()
+        self.main_model = model = main.model()
         msg = model.message_updated
         model.add_observer(msg, self.emit_update)
 
     def gather_matches(self, case_sensitive):
-        if case_sensitive:
-            transform = lambda x: x
-            compare = self.completion_cmp
-        else:
-            transform = lambda x: x.lower()
-            compare = self.lower_completion_cmp
-
-        matched_text = self.matched_text
-        if matched_text:
-            matched_refs = [r for r in self.matches()
-                            if transform(matched_text) in transform(r)]
-            # if we match nothing, still offer to complete something
-            if not matched_refs:
-                matched_refs = self.matches()
-        else:
-            matched_refs = self.matches()
-
-        matched_refs.sort(cmp=compare)
-        return (matched_refs, (), set())
+        refs = filter_matches(self.match_text, self.matches(), case_sensitive)
+        return (refs, (), set())
 
     def emit_update(self):
-        self.emit(SIGNAL('update()'))
+        try:
+            self.emit(SIGNAL(UPDATE_SIGNAL))
+        except RuntimeError: # C++ object has been deleted
+            self.dispose()
 
     def matches(self):
         return []
 
     def dispose(self):
-        self.cola_model.remove_observer(self.emit_update)
+        self.main_model.remove_observer(self.emit_update)
 
 
 class GitRefCompletionModel(GitCompletionModel):
@@ -404,7 +377,7 @@ class GitRefCompletionModel(GitCompletionModel):
         GitCompletionModel.__init__(self, parent)
 
     def matches(self):
-        model = self.cola_model
+        model = self.main_model
         return model.local_branches + model.remote_branches + model.tags
 
 
@@ -415,7 +388,7 @@ class GitBranchCompletionModel(GitCompletionModel):
         GitCompletionModel.__init__(self, parent)
 
     def matches(self):
-        model = self.cola_model
+        model = self.main_model
         return model.local_branches
 
 
@@ -426,42 +399,65 @@ class GitRemoteBranchCompletionModel(GitCompletionModel):
         GitCompletionModel.__init__(self, parent)
 
     def matches(self):
-        model = self.cola_model
+        model = self.main_model
         return model.remote_branches
 
 
+class GitPathCompletionModel(GitCompletionModel):
+    """Base class for path completion"""
+
+    def __init__(self, parent):
+        GitCompletionModel.__init__(self, parent)
+
+    def candidate_paths(self):
+        return []
+
+    def gather_matches(self, case_sensitive):
+        paths, dirs = filter_path_matches(self.match_text,
+                                          self.candidate_paths(),
+                                          case_sensitive)
+        return ((), paths, dirs)
+
+
+class GitStatusFilterCompletionModel(GitPathCompletionModel):
+    """Completer for modified files and folders for status filtering"""
+
+    def __init__(self, parent):
+        GitPathCompletionModel.__init__(self, parent)
+
+    def candidate_paths(self):
+        model = self.main_model
+        return (model.staged + model.unmerged +
+                model.modified + model.untracked)
+
+
 class GitLogCompletionModel(GitRefCompletionModel):
     """Completer for arguments suitable for git-log like commands"""
 
     def __init__(self, parent):
         GitRefCompletionModel.__init__(self, parent)
+        self.connect(self, SIGNAL(UPDATE_SIGNAL), self.gather_paths)
+        self._paths = []
+        self._updated = False
+
+    def gather_paths(self):
+        self._paths = gitcmds.tracked_files()
 
     def gather_matches(self, case_sensitive):
-        (matched_refs, dummy_paths, dummy_dirs) =\
-                GitRefCompletionModel.gather_matches(self, case_sensitive)
+        if not self._paths:
+            self.gather_paths()
+        refs = filter_matches(self.match_text, self.matches(), case_sensitive)
+        paths, dirs = filter_path_matches(self.match_text, self._paths,
+                                          case_sensitive)
+        has_doubledash = (self.match_text == '--' or
+                          self.full_text.startswith('-- ') or
+                          ' -- ' in self.full_text)
+        if has_doubledash:
+            refs = []
+        elif refs and paths:
+            paths.insert(0, '--')
 
-        file_list = self.cola_model.everything()
-        files = set(file_list)
-        files_and_dirs = utils.add_parents(set(files))
-
-        if case_sensitive:
-            transform = lambda x: x
-            compare = self.completion_cmp
-        else:
-            transform = lambda x: x.lower()
-            compare = self.lower_completion_cmp
-
-        dirs = files_and_dirs.difference(files)
-        matched_text = self.matched_text
-        if matched_text:
-            matched_paths = [f for f in files_and_dirs
-                             if transform(matched_text) in transform(f)]
-        else:
-            matched_paths = list(files_and_dirs)
-
-        matched_paths.sort(cmp=compare)
-
-        return (matched_refs, matched_paths, dirs)
+        return (refs, paths, dirs)
 
 
 def bind_lineedit(model):
@@ -469,16 +465,19 @@ def bind_lineedit(model):
 
     class BoundLineEdit(CompletionLineEdit):
 
-        def __init__(self, parent=None):
-            CompletionLineEdit.__init__(self, model, parent)
+        def __init__(self, hint='', parent=None):
+            CompletionLineEdit.__init__(self, model,
+                                        hint=hint, parent=parent)
 
     return BoundLineEdit
 
+
 # Concrete classes
 GitLogLineEdit = bind_lineedit(GitLogCompletionModel)
 GitRefLineEdit = bind_lineedit(GitRefCompletionModel)
 GitBranchLineEdit = bind_lineedit(GitBranchCompletionModel)
 GitRemoteBranchLineEdit = bind_lineedit(GitRemoteBranchCompletionModel)
+GitStatusFilterLineEdit = bind_lineedit(GitStatusFilterCompletionModel)
 
 
 class GitDialog(QtGui.QDialog):
@@ -491,7 +490,7 @@ class GitDialog(QtGui.QDialog):
         self.label = QtGui.QLabel()
         self.label.setText(title)
 
-        self.lineedit = lineedit(self)
+        self.lineedit = lineedit()
         self.setFocusProxy(self.lineedit)
 
         self.ok_button = QtGui.QPushButton()
@@ -501,20 +500,13 @@ class GitDialog(QtGui.QDialog):
         self.close_button = QtGui.QPushButton()
         self.close_button.setText(N_('Close'))
 
-        self.button_layout = QtGui.QHBoxLayout()
-        self.button_layout.setMargin(defs.no_margin)
-        self.button_layout.setSpacing(defs.button_spacing)
-        self.button_layout.addStretch()
-        self.button_layout.addWidget(self.ok_button)
-        self.button_layout.addWidget(self.close_button)
+        self.button_layout = qtutils.hbox(defs.no_margin, defs.button_spacing,
+                                          qtutils.STRETCH,
+                                          self.ok_button, self.close_button)
 
-        self.main_layout = QtGui.QVBoxLayout()
-        self.main_layout.setMargin(defs.margin)
-        self.main_layout.setSpacing(defs.spacing)
-
-        self.main_layout.addWidget(self.label)
-        self.main_layout.addWidget(self.lineedit)
-        self.main_layout.addLayout(self.button_layout)
+        self.main_layout = qtutils.vbox(defs.margin, defs.spacing,
+                                        self.label, self.lineedit,
+                                        self.button_layout)
         self.setLayout(self.main_layout)
 
         qtutils.connect_button(self.ok_button, self.accept)
@@ -522,12 +514,13 @@ class GitDialog(QtGui.QDialog):
 
         self.connect(self.lineedit, SIGNAL('textChanged(const QString&)'),
                      self.text_changed)
+        self.connect(self.lineedit, SIGNAL('returnPressed()'), self.accept)
 
         self.setWindowModality(Qt.WindowModal)
         self.ok_button.setEnabled(False)
 
     def text(self):
-        return unicode(self.lineedit.text())
+        return ustr(self.lineedit.text())
 
     def text_changed(self, txt):
         self.ok_button.setEnabled(bool(self.text()))
