@@ -1,22 +1,30 @@
+from __future__ import division, absolute_import, unicode_literals
+
 from PyQt4 import QtCore
 from PyQt4 import QtGui
 from PyQt4.QtCore import Qt, SIGNAL
 
-import cola
 from cola import cmds
 from cola import core
 from cola import gitcmds
 from cola import gravatar
 from cola import qtutils
 from cola.cmds import run
-from cola.compat import set
 from cola.i18n import N_
+from cola.models import main
+from cola.models import selection
+from cola.qtutils import add_action
+from cola.qtutils import create_action_button
+from cola.qtutils import create_menu
 from cola.qtutils import DiffSyntaxHighlighter
+from cola.qtutils import options_icon
 from cola.widgets import defs
 from cola.widgets.text import MonoTextView
+from cola.compat import ustr
 
 
 COMMITS_SELECTED = 'COMMITS_SELECTED'
+FILES_SELECTED = 'FILES_SELECTED'
 
 
 class DiffTextEdit(MonoTextView):
@@ -27,21 +35,65 @@ class DiffTextEdit(MonoTextView):
         self.highlighter = DiffSyntaxHighlighter(self.document(),
                                                  whitespace=whitespace)
 
+class DiffEditorWidget(QtGui.QWidget):
+
+    def __init__(self, parent=None):
+        QtGui.QWidget.__init__(self, parent)
+
+        self.editor = DiffEditor(self, parent.titleBarWidget())
+        self.main_layout = qtutils.vbox(defs.no_margin, defs.spacing,
+                                        self.editor)
+        self.setLayout(self.main_layout)
+
 
 class DiffEditor(DiffTextEdit):
 
-    def __init__(self, parent):
+    def __init__(self, parent, titlebar):
         DiffTextEdit.__init__(self, parent)
-        self.model = model = cola.model()
-        self.mode = self.model.mode_none
+        self.model = model = main.model()
 
-        self.action_process_section = qtutils.add_action(self,
-                N_('Process Section'),
-                self.apply_section, Qt.Key_H)
-        self.action_process_selection = qtutils.add_action(self,
-                N_('Process Selection'),
+        # "Diff Options" tool menu
+        self.diff_ignore_space_at_eol_action = add_action(self,
+                N_('Ignore changes in whitespace at EOL'),
+                self._update_diff_opts)
+        self.diff_ignore_space_at_eol_action.setCheckable(True)
+
+        self.diff_ignore_space_change_action = add_action(self,
+                N_('Ignore changes in amount of whitespace'),
+                self._update_diff_opts)
+        self.diff_ignore_space_change_action.setCheckable(True)
+
+        self.diff_ignore_all_space_action = add_action(self,
+                N_('Ignore all whitespace'),
+                self._update_diff_opts)
+        self.diff_ignore_all_space_action.setCheckable(True)
+
+        self.diff_function_context_action = add_action(self,
+                N_('Show whole surrounding functions of changes'),
+                self._update_diff_opts)
+        self.diff_function_context_action.setCheckable(True)
+
+        self.diffopts_button = create_action_button(
+                tooltip=N_('Diff Options'), icon=options_icon())
+        self.diffopts_menu = create_menu(N_('Diff Options'),
+                                         self.diffopts_button)
+
+        self.diffopts_menu.addAction(self.diff_ignore_space_at_eol_action)
+        self.diffopts_menu.addAction(self.diff_ignore_space_change_action)
+        self.diffopts_menu.addAction(self.diff_ignore_all_space_action)
+        self.diffopts_menu.addAction(self.diff_function_context_action)
+        self.diffopts_button.setMenu(self.diffopts_menu)
+        qtutils.hide_button_menu_indicator(self.diffopts_button)
+
+        titlebar.add_corner_widget(self.diffopts_button)
+
+        self.action_apply_selection = qtutils.add_action(self, '',
                 self.apply_selection, Qt.Key_S)
 
+        self.action_revert_selection = qtutils.add_action(self, '',
+                self.revert_selection)
+        self.action_revert_selection.setIcon(qtutils.icon('undo.svg'))
+
         self.launch_editor = qtutils.add_action(self,
                 cmds.LaunchEditor.name(), run(cmds.LaunchEditor),
                 cmds.LaunchEditor.SHORTCUT,
@@ -51,50 +103,36 @@ class DiffEditor(DiffTextEdit):
         self.launch_difftool = qtutils.add_action(self,
                 cmds.LaunchDifftool.name(), run(cmds.LaunchDifftool),
                 cmds.LaunchDifftool.SHORTCUT)
-        self.launch_difftool.setIcon(qtutils.icon('git.svg'))
+        self.launch_difftool.setIcon(qtutils.git_icon())
 
-        self.action_stage_selection = qtutils.add_action(self,
-                N_('Stage &Selected Lines'),
-                self.stage_selection)
-        self.action_stage_selection.setIcon(qtutils.icon('add.svg'))
-        self.action_stage_selection.setShortcut(Qt.Key_S)
-
-        self.action_revert_selection = qtutils.add_action(self,
-                N_('Revert Selected Lines...'),
-                self.revert_selection)
-        self.action_revert_selection.setIcon(qtutils.icon('undo.svg'))
-
-        self.action_unstage_selection = qtutils.add_action(self,
-                N_('Unstage &Selected Lines'),
-                self.unstage_selection)
-        self.action_unstage_selection.setIcon(qtutils.icon('remove.svg'))
-        self.action_unstage_selection.setShortcut(Qt.Key_S)
-
-        self.action_apply_selection = qtutils.add_action(self,
-                N_('Apply Diff Selection to Work Tree'),
-                self.stage_selection)
-        self.action_apply_selection.setIcon(qtutils.apply_icon())
-
-        model.add_observer(model.message_mode_about_to_change,
-                           self._mode_about_to_change)
         model.add_observer(model.message_diff_text_changed, self._emit_text)
 
-        self.connect(self, SIGNAL('copyAvailable(bool)'),
-                     self.enable_selection_actions)
-
         self.connect(self, SIGNAL('set_text'), self.setPlainText)
 
     def _emit_text(self, text):
         self.emit(SIGNAL('set_text'), text)
 
+    def _update_diff_opts(self):
+        space_at_eol = self.diff_ignore_space_at_eol_action.isChecked()
+        space_change = self.diff_ignore_space_change_action.isChecked()
+        all_space = self.diff_ignore_all_space_action.isChecked()
+        function_context = self.diff_function_context_action.isChecked()
+
+        gitcmds.update_diff_overrides(space_at_eol,
+                                      space_change,
+                                      all_space,
+                                      function_context)
+        self.emit(SIGNAL('diff_options_updated()'))
+
     # Qt overrides
     def contextMenuEvent(self, event):
         """Create the context menu for the diff display."""
         menu = QtGui.QMenu(self)
-        s = cola.selection()
+        s = selection.selection()
+        filename = selection.filename()
 
-        if self.model.stageable():
-            if s.modified and s.modified[0] in cola.model().submodules:
+        if self.model.stageable() and s.modified:
+            if s.modified[0] in main.model().submodules:
                 action = menu.addAction(qtutils.icon('add.svg'),
                                         cmds.Stage.name(),
                                         cmds.run(cmds.Stage, s.modified))
@@ -103,20 +141,24 @@ class DiffEditor(DiffTextEdit):
                                N_('Launch git-cola'),
                                cmds.run(cmds.OpenRepo,
                                         core.abspath(s.modified[0])))
-            elif s.modified:
-                action = menu.addAction(qtutils.icon('add.svg'),
-                                        N_('Stage Section'),
-                                        self.stage_section)
-                action.setShortcut(Qt.Key_H)
-                menu.addAction(self.action_stage_selection)
-                menu.addSeparator()
-                menu.addAction(qtutils.icon('undo.svg'),
-                               N_('Revert Section...'),
-                               self.revert_section)
+            else:
+                if self.has_selection():
+                    apply_text = N_('Stage Selected Lines')
+                    revert_text = N_('Revert Selected Lines...')
+                else:
+                    apply_text = N_('Stage Diff Hunk')
+                    revert_text = N_('Revert Diff Hunk...')
+
+                self.action_apply_selection.setText(apply_text)
+                self.action_apply_selection.setIcon(qtutils.icon('add.svg'))
+
+                self.action_revert_selection.setText(revert_text)
+
+                menu.addAction(self.action_apply_selection)
                 menu.addAction(self.action_revert_selection)
 
         if self.model.unstageable():
-            if s.staged and s.staged[0] in cola.model().submodules:
+            if s.staged and s.staged[0] in main.model().submodules:
                 action = menu.addAction(qtutils.icon('remove.svg'),
                                         cmds.Unstage.name(),
                                         cmds.do(cmds.Unstage, s.staged))
@@ -126,15 +168,24 @@ class DiffEditor(DiffTextEdit):
                                cmds.do(cmds.OpenRepo,
                                        core.abspath(s.staged[0])))
             elif s.staged:
-                action = menu.addAction(qtutils.icon('remove.svg'),
-                                        N_('Unstage Section'),
-                                        self.unstage_section)
-                action.setShortcut(Qt.Key_H)
-                menu.addAction(self.action_unstage_selection)
+                if self.has_selection():
+                    apply_text = N_('Unstage Selected Lines')
+                else:
+                    apply_text = N_('Unstage Diff Hunk')
+
+                self.action_apply_selection.setText(apply_text)
+                self.action_apply_selection.setIcon(qtutils.icon('remove.svg'))
+
+                menu.addAction(self.action_apply_selection)
 
         if self.model.stageable() or self.model.unstageable():
-            menu.addSeparator()
-            menu.addAction(self.launch_editor)
+            # Do not show the "edit" action when the file does not exist.
+            # Untracked files exist by definition.
+            if filename and core.exists(filename):
+                menu.addSeparator()
+                menu.addAction(self.launch_editor)
+
+            # Removed files can still be diffed.
             menu.addAction(self.launch_difftool)
 
         menu.addSeparator()
@@ -164,20 +215,17 @@ class DiffEditor(DiffTextEdit):
             # Intercept right-click to move the cursor to the current position.
             # setTextCursor() clears the selection so this is only done when
             # nothing is selected.
-            _, selection = self.offset_and_selection()
-            if not selection:
+            if not self.has_selection():
                 cursor = self.cursorForPosition(event.pos())
                 self.setTextCursor(cursor)
 
         return DiffTextEdit.mousePressEvent(self, event)
 
-    def _mode_about_to_change(self, mode):
-        self.mode = mode
-
     def setPlainText(self, text):
         """setPlainText(str) while retaining scrollbar positions"""
-        highlight = (self.mode != self.model.mode_none and
-                     self.mode != self.model.mode_untracked)
+        mode = self.model.mode
+        highlight = (mode != self.model.mode_none and
+                     mode != self.model.mode_untracked)
         self.highlighter.set_enabled(highlight)
 
         scrollbar = self.verticalScrollBar()
@@ -189,18 +237,18 @@ class DiffEditor(DiffTextEdit):
         if text is None:
             return
 
-        offset, selection = self.offset_and_selection()
-        old_text = unicode(self.toPlainText())
+        offset, selection_text = self.offset_and_selection()
+        old_text = ustr(self.toPlainText())
 
         DiffTextEdit.setPlainText(self, text)
 
         # If the old selection exists in the new text then
         # re-select it.
-        if selection and selection in text:
-            idx = text.index(selection)
+        if selection_text and selection_text in text:
+            idx = text.index(selection_text)
             cursor = self.textCursor()
             cursor.setPosition(idx)
-            cursor.setPosition(idx + len(selection),
+            cursor.setPosition(idx + len(selection_text),
                                QtGui.QTextCursor.KeepAnchor)
             self.setTextCursor(cursor)
 
@@ -214,84 +262,50 @@ class DiffEditor(DiffTextEdit):
         if scrollbar and scrollvalue is not None:
             scrollbar.setValue(scrollvalue)
 
+    def has_selection(self):
+        cursor = self.textCursor()
+        return cursor.selectionStart() != cursor.selectionEnd()
+
     def offset_and_selection(self):
         cursor = self.textCursor()
-        offset = cursor.position()
-        selection = unicode(cursor.selection().toPlainText())
-        return offset, selection
-
-    # Mutators
-    def enable_selection_actions(self, enabled):
-        self.action_apply_selection.setEnabled(enabled)
-        self.action_revert_selection.setEnabled(enabled)
-        self.action_unstage_selection.setEnabled(enabled)
-        self.action_stage_selection.setEnabled(enabled)
-
-    def apply_section(self):
-        s = cola.single_selection()
-        if self.model.stageable() and s.modified:
-            self.stage_section()
-        elif self.model.unstageable():
-            self.unstage_section()
+        offset = cursor.selectionStart()
+        selection_text = ustr(cursor.selection().toPlainText())
+        return offset, selection_text
 
     def apply_selection(self):
-        s = cola.single_selection()
+        s = selection.single_selection()
         if self.model.stageable() and s.modified:
-            self.stage_selection()
+            self.process_diff_selection(staged=False)
         elif self.model.unstageable():
-            self.unstage_selection()
-
-    def stage_section(self):
-        """Stage a specific section."""
-        self.process_diff_selection(staged=False)
-
-    def stage_selection(self):
-        """Stage selected lines."""
-        self.process_diff_selection(staged=False, selected=True)
-
-    def unstage_section(self, cached=True):
-        """Unstage a section."""
-        self.process_diff_selection(staged=True)
-
-    def unstage_selection(self):
-        """Unstage selected lines."""
-        self.process_diff_selection(staged=True, selected=True)
-
-    def revert_section(self):
-        """Destructively remove a section from a worktree file."""
-        if not qtutils.confirm(N_('Revert Section?'),
-                               N_('This operation drops uncommitted changes.\n'
-                                  'These changes cannot be recovered.'),
-                               N_('Revert the uncommitted changes?'),
-                               N_('Revert Section'),
-                               default=True,
-                               icon=qtutils.icon('undo.svg')):
-            return
-        self.process_diff_selection(staged=False, apply_to_worktree=True,
-                                    reverse=True)
+            self.process_diff_selection(staged=True)
 
     def revert_selection(self):
-        """Destructively check out content for the selected file from $head."""
-        if not qtutils.confirm(N_('Revert Selected Lines?'),
+        """Destructively revert selected lines or hunk from a worktree file."""
+
+        if self.has_selection():
+            title = N_('Revert Selected Lines?')
+            ok_text = N_('Revert Selected Lines')
+        else:
+            title = N_('Revert Diff Hunk?')
+            ok_text = N_('Revert Diff Hunk')
+
+        if not qtutils.confirm(title,
                                N_('This operation drops uncommitted changes.\n'
                                   'These changes cannot be recovered.'),
                                N_('Revert the uncommitted changes?'),
-                               N_('Revert Selected Lines'),
+                               ok_text,
                                default=True,
                                icon=qtutils.icon('undo.svg')):
             return
-        self.process_diff_selection(staged=False, apply_to_worktree=True,
-                                    reverse=True, selected=True)
+        self.process_diff_selection(staged=False, apply_to_worktree=True)
 
-    def process_diff_selection(self, selected=False,
-                               staged=True, apply_to_worktree=False,
-                               reverse=False):
+    def process_diff_selection(self, staged=True, apply_to_worktree=False):
         """Implement un/staging of selected lines or sections."""
-        offset, selection = self.offset_and_selection()
-        if cola.selection_model().is_empty():
+        if selection.selection_model().is_empty():
             return
+        offset, selection_text = self.offset_and_selection()
         cmds.do(cmds.ApplyDiffSelection,
-                staged, selected, offset, selection, apply_to_worktree)
+                staged, offset, selection_text, apply_to_worktree)
 
 
 
@@ -306,7 +320,7 @@ class DiffWidget(QtGui.QWidget):
         summary_font = QtGui.QFont(author_font)
         summary_font.setWeight(QtGui.QFont.Bold)
 
-        policy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding,
+        policy = QtGui.QSizePolicy(QtGui.QSizePolicy.MinimumExpanding,
                                    QtGui.QSizePolicy.Minimum)
 
         self.gravatar_label = gravatar.GravatarLabel()
@@ -335,27 +349,20 @@ class DiffWidget(QtGui.QWidget):
         self.tasks = set()
         self.reflector = QtCore.QObject(self)
 
-        self.info_layout = QtGui.QVBoxLayout()
-        self.info_layout.setMargin(defs.no_margin)
-        self.info_layout.setSpacing(defs.no_spacing)
-        self.info_layout.addWidget(self.author_label)
-        self.info_layout.addWidget(self.summary_label)
-        self.info_layout.addWidget(self.sha1_label)
+        self.info_layout = qtutils.vbox(defs.no_margin, defs.no_spacing,
+                                        self.author_label, self.summary_label,
+                                        self.sha1_label)
 
-        self.logo_layout = QtGui.QHBoxLayout()
+        self.logo_layout = qtutils.hbox(defs.no_margin, defs.button_spacing,
+                                        self.gravatar_label, self.info_layout)
         self.logo_layout.setContentsMargins(defs.margin, 0, defs.margin, 0)
-        self.logo_layout.setSpacing(defs.button_spacing)
-        self.logo_layout.addWidget(self.gravatar_label)
-        self.logo_layout.addLayout(self.info_layout)
 
-        self.main_layout = QtGui.QVBoxLayout()
-        self.main_layout.setMargin(defs.no_margin)
-        self.main_layout.setSpacing(defs.spacing)
-        self.main_layout.addLayout(self.logo_layout)
-        self.main_layout.addWidget(self.diff)
+        self.main_layout = qtutils.vbox(defs.no_margin, defs.spacing,
+                                        self.logo_layout, self.diff)
         self.setLayout(self.main_layout)
 
         notifier.add_observer(COMMITS_SELECTED, self.commits_selected)
+        notifier.add_observer(FILES_SELECTED, self.files_selected)
         self.connect(self.reflector, SIGNAL('diff'), self.diff.setText)
         self.connect(self.reflector, SIGNAL('task_done'), self.task_done)
 
@@ -365,9 +372,9 @@ class DiffWidget(QtGui.QWidget):
         except:
             pass
 
-    def set_diff_sha1(self, sha1):
+    def set_diff_sha1(self, sha1, filename=None):
         self.diff.setText('+++ ' + N_('Loading...'))
-        task = DiffInfoTask(sha1, self.reflector)
+        task = DiffInfoTask(sha1, self.reflector, filename=filename)
         self.tasks.add(task)
         QtCore.QThreadPool.globalInstance().start(task)
 
@@ -375,7 +382,7 @@ class DiffWidget(QtGui.QWidget):
         if len(commits) != 1:
             return
         commit = commits[0]
-        sha1 = commit.sha1
+        self.sha1 = commit.sha1
 
         email = commit.email or ''
         summary = commit.summary or ''
@@ -395,13 +402,19 @@ class DiffWidget(QtGui.QWidget):
         author_template = '%(author)s <%(email)s>' % template_args
         self.author_label.set_template(author_text, author_template)
         self.summary_label.set_text(summary)
-        self.sha1_label.set_text(sha1)
+        self.sha1_label.set_text(self.sha1)
 
-        self.set_diff_sha1(sha1)
+        self.set_diff_sha1(self.sha1)
         self.gravatar_label.set_email(email)
 
+    def files_selected(self, filenames):
+        if not filenames:
+            return
+        self.set_diff_sha1(self.sha1, filenames[0])
+
 
 class TextLabel(QtGui.QLabel):
+
     def __init__(self, parent=None):
         QtGui.QLabel.__init__(self, parent)
         self.setTextInteractionFlags(Qt.TextSelectableByMouse |
@@ -432,7 +445,7 @@ class TextLabel(QtGui.QLabel):
             return
         text = self._metrics.elidedText(self._template,
                                         Qt.ElideRight, width-2)
-        if unicode(text) != self._template:
+        if ustr(text) != self._template:
             self._display = text
 
     # Qt overrides
@@ -450,11 +463,13 @@ class TextLabel(QtGui.QLabel):
 
 
 class DiffInfoTask(QtCore.QRunnable):
-    def __init__(self, sha1, reflector):
+
+    def __init__(self, sha1, reflector, filename=None):
         QtCore.QRunnable.__init__(self)
         self.sha1 = sha1
         self.reflector = reflector
+        self.filename = filename
 
     def run(self):
-        diff = gitcmds.diff_info(self.sha1)
+        diff = gitcmds.diff_info(self.sha1, filename=self.filename)
         self.reflector.emit(SIGNAL('diff'), diff)
