@@ -1,5 +1,5 @@
-import os
-import subprocess
+from __future__ import division, absolute_import, unicode_literals
+
 import itertools
 
 from PyQt4 import QtCore
@@ -7,15 +7,15 @@ from PyQt4 import QtGui
 from PyQt4.QtCore import Qt
 from PyQt4.QtCore import SIGNAL
 
-import cola
 from cola import cmds
 from cola import core
 from cola import qtutils
 from cola import utils
-from cola.compat import set
 from cola.i18n import N_
-from cola.interaction import Interaction
-from cola.models.selection import State
+from cola.models import main
+from cola.models import selection
+from cola.widgets import completion
+from cola.widgets import defs
 
 
 class StatusWidget(QtGui.QWidget):
@@ -26,14 +26,35 @@ class StatusWidget(QtGui.QWidget):
     Qt signals.
 
     """
-    def __init__(self, parent=None):
+    def __init__(self, titlebar, parent=None):
         QtGui.QWidget.__init__(self, parent)
-        self.layout = QtGui.QVBoxLayout(self)
-        self.setLayout(self.layout)
 
-        self.tree = StatusTreeWidget(self)
-        self.layout.addWidget(self.tree)
-        self.layout.setContentsMargins(0, 0, 0, 0)
+        tooltip = N_('Toggle the paths filter')
+        self.filter_button = qtutils.create_action_button(
+                tooltip=tooltip,
+                icon=qtutils.filter_icon())
+
+        self.filter_widget = StatusFilterWidget()
+        self.filter_widget.hide()
+        self.tree = StatusTreeWidget()
+
+        self.main_layout = qtutils.vbox(defs.no_margin, defs.no_spacing,
+                                        self.filter_widget, self.tree)
+        self.setLayout(self.main_layout)
+
+        self.toggle_action = qtutils.add_action(self, tooltip,
+                self.toggle_filter, 'Shift+Ctrl+F')
+
+        titlebar.add_corner_widget(self.filter_button)
+        qtutils.connect_button(self.filter_button, self.toggle_filter)
+
+    def toggle_filter(self):
+        shown = not self.filter_widget.isVisible()
+        self.filter_widget.setVisible(shown)
+        if shown:
+            self.filter_widget.setFocus(True)
+        else:
+            self.tree.setFocus(True)
 
     def set_initial_size(self):
         self.setMaximumWidth(222)
@@ -45,6 +66,11 @@ class StatusWidget(QtGui.QWidget):
     def refresh(self):
         self.tree.show_selection()
 
+    def set_filter(self, txt):
+        self.filter_widget.setVisible(True)
+        self.filter_widget.text.set_value(txt)
+        self.filter_widget.apply_filter()
+
 
 class StatusTreeWidget(QtGui.QTreeWidget):
     # Item categories
@@ -58,7 +84,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
     # Read-only access to the mode state
     mode = property(lambda self: self.m.mode)
 
-    def __init__(self, parent):
+    def __init__(self, parent=None):
         QtGui.QTreeWidget.__init__(self, parent)
 
         self.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
@@ -69,6 +95,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         self.setAnimated(True)
         self.setRootIsDecorated(False)
         self.setIndentation(0)
+        self.setDragEnabled(True)
 
         self.add_item(N_('Staged'), hide=True)
         self.add_item(N_('Unmerged'), hide=True)
@@ -82,22 +109,35 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         self.old_current_item = None
         self.expanded_items = set()
 
-        self.process_selection = qtutils.add_action(self,
-                N_('Stage / Unstage'), self._process_selection,
-                cmds.Stage.SHORTCUT)
+        self.process_selection_action = qtutils.add_action(self,
+                cmds.StageOrUnstage.name(),
+                cmds.run(cmds.StageOrUnstage),
+                cmds.StageOrUnstage.SHORTCUT)
 
-        self.launch_difftool = qtutils.add_action(self,
+        self.revert_unstaged_edits_action = qtutils.add_action(self,
+                cmds.RevertUnstagedEdits.name(),
+                cmds.run(cmds.RevertUnstagedEdits),
+                cmds.RevertUnstagedEdits.SHORTCUT)
+        self.revert_unstaged_edits_action.setIcon(qtutils.icon('undo.svg'))
+
+        self.revert_uncommitted_edits_action = qtutils.add_action(self,
+                cmds.RevertUncommittedEdits.name(),
+                cmds.run(cmds.RevertUncommittedEdits),
+                cmds.RevertUncommittedEdits.SHORTCUT)
+        self.revert_uncommitted_edits_action.setIcon(qtutils.icon('undo.svg'))
+
+        self.launch_difftool_action = qtutils.add_action(self,
                 cmds.LaunchDifftool.name(),
                 cmds.run(cmds.LaunchDifftool),
                 cmds.LaunchDifftool.SHORTCUT)
-        self.launch_difftool.setIcon(qtutils.icon('git.svg'))
+        self.launch_difftool_action.setIcon(qtutils.git_icon())
 
-        self.launch_editor = qtutils.add_action(self,
+        self.launch_editor_action = qtutils.add_action(self,
                 cmds.LaunchEditor.name(),
                 cmds.run(cmds.LaunchEditor),
                 cmds.LaunchEditor.SHORTCUT,
                 'Return', 'Enter')
-        self.launch_editor.setIcon(qtutils.options_icon())
+        self.launch_editor_action.setIcon(qtutils.options_icon())
 
         if not utils.is_win32():
             self.open_using_default_app = qtutils.add_action(self,
@@ -106,27 +146,47 @@ class StatusTreeWidget(QtGui.QTreeWidget):
                     cmds.OpenDefaultApp.SHORTCUT)
             self.open_using_default_app.setIcon(qtutils.file_icon())
 
-            self.open_parent_dir = qtutils.add_action(self,
+            self.open_parent_dir_action = qtutils.add_action(self,
                     cmds.OpenParentDir.name(),
                     self._open_parent_dir,
                     cmds.OpenParentDir.SHORTCUT)
-            self.open_parent_dir.setIcon(qtutils.open_file_icon())
+            self.open_parent_dir_action.setIcon(qtutils.open_file_icon())
 
-        self.up = qtutils.add_action(self, N_('Move Up'), self.move_up,
-                                     Qt.Key_K)
+        self.up_action = qtutils.add_action(self,
+                N_('Move Up'), self.move_up, Qt.Key_K)
 
-        self.down = qtutils.add_action(self, N_('Move Down'), self.move_down,
-                                       Qt.Key_J)
+        self.down_action = qtutils.add_action(self,
+                N_('Move Down'), self.move_down, Qt.Key_J)
 
         self.copy_path_action = qtutils.add_action(self,
                 N_('Copy Path to Clipboard'),
                 self.copy_path, QtGui.QKeySequence.Copy)
         self.copy_path_action.setIcon(qtutils.theme_icon('edit-copy.svg'))
 
+        self.copy_relpath_action = qtutils.add_action(self,
+                N_('Copy Relative Path to Clipboard'),
+                self.copy_relpath, QtGui.QKeySequence.Cut)
+        self.copy_relpath_action.setIcon(qtutils.theme_icon('edit-copy.svg'))
+
+        if cmds.MoveToTrash.AVAILABLE:
+            self.move_to_trash_action = qtutils.add_action(self,
+                    N_('Move file(s) to trash'),
+                    self._trash_untracked_files, cmds.MoveToTrash.SHORTCUT)
+            self.move_to_trash_action.setIcon(qtutils.discard_icon())
+            delete_shortcut = cmds.Delete.SHORTCUT
+        else:
+            self.move_to_trash_action = None
+            delete_shortcut = cmds.Delete.ALT_SHORTCUT
+
+        self.delete_untracked_files_action = qtutils.add_action(self,
+                N_('Delete File(s)...'),
+                self._delete_untracked_files, delete_shortcut)
+        self.delete_untracked_files_action.setIcon(qtutils.discard_icon())
+
         self.connect(self, SIGNAL('about_to_update'), self._about_to_update)
         self.connect(self, SIGNAL('updated'), self._updated)
 
-        self.m = cola.model()
+        self.m = main.model()
         self.m.add_observer(self.m.message_about_to_update,
                             self.about_to_update)
         self.m.add_observer(self.m.message_updated, self.updated)
@@ -148,7 +208,6 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         # TODO no icon
         font = self.font()
         font.setBold(True)
-        font.setCapitalization(QtGui.QFont.SmallCaps)
 
         item = QtGui.QTreeWidgetItem(self)
         item.setFont(0, font)
@@ -217,13 +276,13 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         # we do not need to rerun the callbacks which were triggered
         # above.  Block signals to skip the callbacks.
         self.blockSignals(True)
-        for (new, old, selection, reselect) in saved_selection:
-            for item in selection:
+        for (new, old, sel, reselect) in saved_selection:
+            for item in sel:
                 if item in new:
                     reselect(item, current=False)
         self.blockSignals(False)
 
-        for (new, old, selection, reselect) in saved_selection:
+        for (new, old, sel, reselect) in saved_selection:
             # When modified is staged, select the next modified item
             # When unmerged is staged, select the next unmerged item
             # When unstaging, select the next staged item
@@ -231,7 +290,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
             if len(new) >= len(old):
                 # The list did not shrink so it is not one of these cases.
                 continue
-            for item in selection:
+            for item in sel:
                 # The item still exists so ignore it
                 if item in new or item not in old:
                     continue
@@ -330,6 +389,14 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         self.restore_selection()
         self.restore_scrollbar()
         self.update_column_widths()
+        self.update_actions()
+
+    def update_actions(self, selected=None):
+        if selected is None:
+            selected = selection.selection()
+        can_revert_edits = bool(selected.staged or selected.modified)
+        self.revert_unstaged_edits_action.setEnabled(can_revert_edits)
+        self.revert_uncommitted_edits_action.setEnabled(can_revert_edits)
 
     def set_staged(self, items):
         """Adds items to the 'Staged' subtree."""
@@ -403,9 +470,9 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         s = self.selection()
         menu = QtGui.QMenu(self)
 
-        selection = self.selected_indexes()
-        if selection:
-            category, idx = selection[0]
+        selected_indexes = self.selected_indexes()
+        if selected_indexes:
+            category, idx = selected_indexes[0]
             # A header item e.g. 'Staged', 'Modified', etc.
             if category == self.idx_header:
                 return self._create_header_context_menu(menu, idx)
@@ -454,10 +521,15 @@ class StatusTreeWidget(QtGui.QTreeWidget):
                                     cmds.run(cmds.Unstage, self.staged()))
             action.setShortcut(cmds.Unstage.SHORTCUT)
 
-        menu.addAction(self.launch_editor)
-        menu.addAction(self.launch_difftool)
+        # Do all of the selected items exist?
+        staged_items = self.staged_items()
+        all_exist = all([i.exists for i in staged_items])
 
-        if not utils.is_win32():
+        if all_exist:
+            menu.addAction(self.launch_editor_action)
+            menu.addAction(self.launch_difftool_action)
+
+        if all_exist and not utils.is_win32():
             menu.addSeparator()
             action = menu.addAction(qtutils.file_icon(),
                     cmds.OpenDefaultApp.name(),
@@ -471,15 +543,11 @@ class StatusTreeWidget(QtGui.QTreeWidget):
 
         if self.m.undoable():
             menu.addSeparator()
-            menu.addAction(qtutils.icon('undo.svg'),
-                           N_('Revert Unstaged Edits...'),
-                           lambda: self._revert_unstaged_edits(staged=True))
-            menu.addAction(qtutils.icon('undo.svg'),
-                           N_('Revert Uncommited Edits...'),
-                           lambda: self._revert_uncommitted_edits(
-                                        self.staged()))
+            menu.addAction(self.revert_unstaged_edits_action)
+
         menu.addSeparator()
         menu.addAction(self.copy_path_action)
+        menu.addAction(self.copy_relpath_action)
         return menu
 
     def _create_staged_submodule_context_menu(self, menu, s):
@@ -488,7 +556,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
                        cmds.run(cmds.OpenRepo,
                                 core.abspath(s.staged[0])))
 
-        menu.addAction(self.launch_editor)
+        menu.addAction(self.launch_editor_action)
         menu.addSeparator()
 
         action = menu.addAction(qtutils.icon('remove.svg'),
@@ -498,17 +566,18 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         menu.addSeparator()
 
         menu.addAction(self.copy_path_action)
+        menu.addAction(self.copy_relpath_action)
         return menu
 
     def _create_unmerged_context_menu(self, menu, s):
-        menu.addAction(self.launch_difftool)
+        menu.addAction(self.launch_difftool_action)
 
         action = menu.addAction(qtutils.icon('add.svg'),
                                 N_('Stage Selected'),
                                 cmds.run(cmds.Stage, self.unstaged()))
         action.setShortcut(cmds.Stage.SHORTCUT)
         menu.addSeparator()
-        menu.addAction(self.launch_editor)
+        menu.addAction(self.launch_editor_action)
 
         if not utils.is_win32():
             menu.addSeparator()
@@ -524,6 +593,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
 
         menu.addSeparator()
         menu.addAction(self.copy_path_action)
+        menu.addAction(self.copy_relpath_action)
         return menu
 
     def _create_unstaged_context_menu(self, menu, s):
@@ -538,24 +608,23 @@ class StatusTreeWidget(QtGui.QTreeWidget):
                                     cmds.run(cmds.Stage, self.unstaged()))
             action.setShortcut(cmds.Stage.SHORTCUT)
 
-        if self.unstaged():
-            menu.addAction(self.launch_editor)
+        # Do all of the selected items exist?
+        unstaged_items = self.unstaged_items()
+        all_exist = all([i.exists for i in unstaged_items])
 
-        if s.modified and self.m.stageable():
-            menu.addAction(self.launch_difftool)
+        if all_exist and self.unstaged():
+            menu.addAction(self.launch_editor_action)
+
+        if all_exist and s.modified and self.m.stageable():
+            menu.addAction(self.launch_difftool_action)
 
         if s.modified and self.m.stageable():
             if self.m.undoable():
                 menu.addSeparator()
-                menu.addAction(qtutils.icon('undo.svg'),
-                               N_('Revert Unstaged Edits...'),
-                               self._revert_unstaged_edits)
-                menu.addAction(qtutils.icon('undo.svg'),
-                               N_('Revert Uncommited Edits...'),
-                               lambda: self._revert_uncommitted_edits(
-                                            self.modified()))
+                menu.addAction(self.revert_unstaged_edits_action)
+                menu.addAction(self.revert_uncommitted_edits_action)
 
-        if self.unstaged() and not utils.is_win32():
+        if all_exist and self.unstaged() and not utils.is_win32():
             menu.addSeparator()
             action = menu.addAction(qtutils.file_icon(),
                     cmds.OpenDefaultApp.name(),
@@ -567,10 +636,11 @@ class StatusTreeWidget(QtGui.QTreeWidget):
                     self._open_parent_dir)
             action.setShortcut(cmds.OpenParentDir.SHORTCUT)
 
-        if s.untracked:
+        if all_exist and s.untracked:
             menu.addSeparator()
-            menu.addAction(qtutils.discard_icon(),
-                           N_('Delete File(s)...'), self._delete_files)
+            if self.move_to_trash_action is not None:
+                menu.addAction(self.move_to_trash_action)
+            menu.addAction(self.delete_untracked_files_action)
             menu.addSeparator()
             menu.addAction(qtutils.icon('edit-clear.svg'),
                            N_('Add to .gitignore'),
@@ -578,6 +648,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
                                 map(lambda x: '/' + x, self.untracked())))
         menu.addSeparator()
         menu.addAction(self.copy_path_action)
+        menu.addAction(self.copy_relpath_action)
         return menu
 
     def _create_modified_submodule_context_menu(self, menu, s):
@@ -585,7 +656,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
                        N_('Launch git-cola'),
                        cmds.run(cmds.OpenRepo, core.abspath(s.modified[0])))
 
-        menu.addAction(self.launch_editor)
+        menu.addAction(self.launch_editor_action)
 
         if self.m.stageable():
             menu.addSeparator()
@@ -596,71 +667,15 @@ class StatusTreeWidget(QtGui.QTreeWidget):
 
         menu.addSeparator()
         menu.addAction(self.copy_path_action)
+        menu.addAction(self.copy_relpath_action)
         return menu
 
 
-    def _delete_files(self):
-        files = self.untracked()
-        count = len(files)
-        if count == 0:
-            return
+    def _delete_untracked_files(self):
+        cmds.do(cmds.Delete, self.untracked())
 
-        title = N_('Delete Files?')
-        msg = N_('The following files will be deleted:') + '\n\n'
-
-        fileinfo = subprocess.list2cmdline(files)
-        if len(fileinfo) > 2048:
-            fileinfo = fileinfo[:2048].rstrip() + '...'
-        msg += fileinfo
-
-        info_txt = N_('Delete %d file(s)?') % count
-        ok_txt = N_('Delete Files')
-
-        if qtutils.confirm(title, msg, info_txt, ok_txt,
-                           default=True,
-                           icon=qtutils.discard_icon()):
-            cmds.do(cmds.Delete, files)
-
-    def _revert_unstaged_edits(self, staged=False):
-        if not self.m.undoable():
-            return
-        if staged:
-            items_to_undo = self.staged()
-        else:
-            items_to_undo = self.modified()
-
-        if items_to_undo:
-            if not qtutils.confirm(N_('Revert Unstaged Changes?'),
-                                   N_('This operation drops unstaged changes.\n'
-                                      'These changes cannot be recovered.'),
-                                   N_('Revert the unstaged changes?'),
-                                   N_('Revert Unstaged Changes'),
-                                   default=True,
-                                   icon=qtutils.icon('undo.svg')):
-                return
-            args = []
-            if not staged and self.m.amending():
-                args.append(self.m.head)
-            cmds.do(cmds.Checkout, args + ['--'] + items_to_undo)
-        else:
-            msg = N_('No files selected for checkout from HEAD.')
-            Interaction.log(msg)
-
-    def _revert_uncommitted_edits(self, items_to_undo):
-        if items_to_undo:
-            if not qtutils.confirm(
-                    N_('Revert Uncommitted Changes?'),
-                    N_('This operation drops uncommitted changes.\n'
-                       'These changes cannot be recovered.'),
-                    N_('Revert the uncommitted changes?'),
-                    N_('Revert Uncommitted Changes'),
-                    default=True,
-                    icon=qtutils.icon('undo.svg')):
-                return
-            cmds.do(cmds.Checkout, [self.m.head, '--'] + items_to_undo)
-        else:
-            msg = N_('No files selected for checkout from HEAD.')
-            Interaction.log(msg)
+    def _trash_untracked_files(self):
+        cmds.do(cmds.MoveToTrash, self.untracked())
 
     def single_selection(self):
         """Scan across staged, modified, etc. and return a single item."""
@@ -679,7 +694,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         elif s.untracked:
             ut = s.untracked[0]
 
-        return State(st, um, m, ut)
+        return selection.State(st, um, m, ut)
 
     def selected_indexes(self):
         """Returns a list of (category, row) representing the tree selection."""
@@ -696,12 +711,12 @@ class StatusTreeWidget(QtGui.QTreeWidget):
 
     def selection(self):
         """Return the current selection in the repo status tree."""
-        return State(self.staged(), self.unmerged(),
-                     self.modified(), self.untracked())
+        return selection.State(self.staged(), self.unmerged(),
+                               self.modified(), self.untracked())
 
     def contents(self):
-        return State(self.m.staged, self.m.unmerged,
-                     self.m.modified, self.m.untracked)
+        return selection.State(self.m.staged, self.m.unmerged,
+                               self.m.modified, self.m.untracked)
 
     def all_files(self):
         c = self.contents()
@@ -709,17 +724,7 @@ class StatusTreeWidget(QtGui.QTreeWidget):
 
     def selected_group(self):
         """A list of selected files in various states of being"""
-        selection = []
-        s = self.selection()
-        if s.staged:
-            selection = s.staged
-        elif s.unmerged:
-            selection = s.unmerged
-        elif s.modified:
-            selection = s.modified
-        elif s.untracked:
-            selection = s.untracked
-        return selection
+        return selection.pick(self.selection())
 
     def selected_idx(self):
         c = self.contents()
@@ -771,50 +776,55 @@ class StatusTreeWidget(QtGui.QTreeWidget):
     def untracked(self):
         return self._subtree_selection(self.idx_untracked, self.m.untracked)
 
+    def staged_items(self):
+        return self._subtree_selection_items(self.idx_staged)
+
+    def unstaged_items(self):
+        return (self.unmerged_items() + self.modified_items() +
+                self.untracked_items())
+
+    def modified_items(self):
+        return self._subtree_selection_items(self.idx_modified)
+
+    def unmerged_items(self):
+        return self._subtree_selection_items(self.idx_unmerged)
+
+    def untracked_items(self):
+        return self._subtree_selection_items(self.idx_untracked)
+
     def _subtree_selection(self, idx, items):
         item = self.topLevelItem(idx)
         return qtutils.tree_selection(item, items)
 
+    def _subtree_selection_items(self, idx):
+        item = self.topLevelItem(idx)
+        return qtutils.tree_selection_items(item)
+
     def double_clicked(self, item, idx):
         """Called when an item is double-clicked in the repo status tree."""
-        self._process_selection()
-
-    def _process_selection(self):
-        s = self.selection()
-        if s.staged:
-            cmds.do(cmds.Unstage, s.staged)
-
-        unstaged = []
-        if s.unmerged:
-            unstaged.extend(s.unmerged)
-        if s.modified:
-            unstaged.extend(s.modified)
-        if s.untracked:
-            unstaged.extend(s.untracked)
-        if unstaged:
-            cmds.do(cmds.Stage, unstaged)
+        cmds.do(cmds.StageOrUnstage)
 
     def _open_using_default_app(self):
-        selection = self.selected_group()
-        cmds.do(cmds.OpenDefaultApp, selection)
+        cmds.do(cmds.OpenDefaultApp, self.selected_group())
 
     def _open_parent_dir(self):
-        selection = self.selected_group()
-        cmds.do(cmds.OpenParentDir, selection)
+        cmds.do(cmds.OpenParentDir, self.selected_group())
 
     def show_selection(self):
         """Show the selected item."""
         # Sync the selection model
-        s = self.selection()
-        cola.selection_model().set_selection(s)
-        selection = self.selected_indexes()
-        if not selection:
+        selected = self.selection()
+        selection.selection_model().set_selection(selected)
+        self.update_actions(selected=selected)
+
+        selected_indexes = self.selected_indexes()
+        if not selected_indexes:
             if self.m.amending():
                 cmds.do(cmds.SetDiffText, '')
             else:
                 cmds.do(cmds.ResetMode)
             return
-        category, idx = selection[0]
+        category, idx = selected_indexes[0]
         # A header item e.g. 'Staged', 'Modified', etc.
         if category == self.idx_header:
             cls = {
@@ -843,9 +853,9 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         idx = self.selected_idx()
         all_files = self.all_files()
         if idx is None:
-            selection = self.selected_indexes()
-            if selection:
-                category, toplevel_idx = selection[0]
+            selected_indexes = self.selected_indexes()
+            if selected_indexes:
+                category, toplevel_idx = selected_indexes[0]
                 if category == self.idx_header:
                     item = self.itemAbove(self.topLevelItem(toplevel_idx))
                     if item is not None:
@@ -863,9 +873,9 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         idx = self.selected_idx()
         all_files = self.all_files()
         if idx is None:
-            selection = self.selected_indexes()
-            if selection:
-                category, toplevel_idx = selection[0]
+            selected_indexes = self.selected_indexes()
+            if selected_indexes:
+                category, toplevel_idx = selected_indexes[0]
                 if category == self.idx_header:
                     item = self.itemBelow(self.topLevelItem(toplevel_idx))
                     if item is not None:
@@ -879,9 +889,43 @@ class StatusTreeWidget(QtGui.QTreeWidget):
         else:
             self.select_by_index(0)
 
-    def copy_path(self):
+    def copy_path(self, absolute=True):
         """Copy a selected path to the clipboard"""
-        filename = cola.selection_model().filename()
-        if filename is not None:
-            curdir = os.getcwdu()
-            qtutils.set_clipboard(os.path.join(curdir, filename))
+        filename = selection.selection_model().filename()
+        qtutils.copy_path(filename, absolute=absolute)
+
+    def copy_relpath(self):
+        """Copy a selected relative path to the clipboard"""
+        self.copy_path(absolute=False)
+
+    def mimeData(self, items):
+        """Return a list of absolute-path URLs"""
+        paths = qtutils.paths_from_items(items, item_filter=lambda x: x.exists)
+        return qtutils.mimedata_from_paths(paths)
+
+    def mimeTypes(self):
+        return qtutils.mimetypes()
+
+
+class StatusFilterWidget(QtGui.QWidget):
+
+    def __init__(self, parent=None):
+        QtGui.QWidget.__init__(self, parent)
+        self.main_model = main.model()
+
+        hint = N_('Filter paths...')
+        self.text = completion.GitStatusFilterLineEdit(hint=hint, parent=self)
+        self.text.setToolTip(hint)
+        self.text.enable_hint(True)
+        self.setFocusProxy(self.text)
+
+        self.main_layout = qtutils.hbox(defs.no_margin, defs.spacing, self.text)
+        self.setLayout(self.main_layout)
+
+        self.connect(self.text, SIGNAL('changed()'), self.apply_filter)
+        self.connect(self.text, SIGNAL('returnPressed()'), self.apply_filter)
+
+    def apply_filter(self):
+        text = self.text.value()
+        paths = utils.shell_split(text)
+        self.main_model.update_path_filter(paths)
